// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_argdispatcher.h,v $
// Copyright(c)'1999-2000 by APACHE Team
// Author: O. Lobry
// $Id: acom_argdispatcher.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================
#ifndef _ACOM_ARGDISPATCHER_H_
#define _ACOM_ARGDISPATCHER_H_

#include <string>
#include <map>
#include <list>
#include <vector>
#include <inuk/acom_properties.h>

namespace ACOM_NET_NAMESPACE {

// ---------------------------------------------------------------------------
//
//! This class aims at dispatching arguments to different modules.
//! Different arguments can be passed to different modules the following way:
//! progname -- module1 [module args] -- module2 [module2 args] 
//!          ... -- modulen [modulen args] -- [application args]
//! To be recognized, module names must registered before parsing
//! Parsing functions can be called several times, and the object can be reset
//! Arguments for a module can be retreived from an object arg of class Arg
//! as a vector of strings (type Arg::ArgVect) as arg["modulename"]. The list 
//! of argument include the program name. Unparsed arguments are put in the 
//! module named "_unparsed_". Arguments can be protected form any module using
//! this keyword (ex: progname -- _unparsed_ [ args not to parse] ... ). 
//! All parsed args can be retreived usong the special module name "_all_".
//! Classical argument infos (int argc, char ** argv) can be created using 
//! the Arg::ArgInfo class before calling the parsing function of a module.
//! @warning Note that when passing ArgInfo fields to a function, the object 
//! must not be destroyed if the function store pointers to arguments as it 
//! could do with standard arguments.
//
// ---------------------------------------------------------------------------

class Arg {
public:

  typedef std::vector<std::string> ArgVect;

  class ArgInfo {
  public:
    //! Default constructor. Create argument infos for a specified module 
    //! from argument object.
    //! @param arg The argument object form which getting information
    //! @param module The module for which gettiong arg infos
    ArgInfo(Arg& arg, const char * module);

    //! Default constructor. Create argument infos from a string.
    //! @param arg a string representing a command line
    ArgInfo(std::string &arg);

    //! Default Dstor
    ~ArgInfo();

  public:
    char ** argv; //! < argument array
    int argc;     //! < argument count
  }; 


  //! Default Constructor. Programe name must be specified at creation time
  //! @param pname The name of the program.
  Arg(const char * pname): _progname(pname) 
    { 
      new_module("_unparsed_"); 
      new_module("_all_"); 
    }

  //! Add a module to recognize by parser
  //! @returns true if the module already exists
  bool new_module(const char * module);
  
  //! Parse arguments for registered modules.
  //! Arguments are added to already parsed ones
  //! @param argc count of arguments
  //! @param argv array of arguments
  void parse(int argc, const char **argv);

  //! Parse arguments for only one module. Others arguments are ignored
  //! and not put in "_unparsed_" special module.
  //! Module is automatically added to module list.
  //! @param argc count of arguments
  //! @param argv array of arguments
  //! @param module name of the concerned module 
  void parse(int argc, const char ** argv, const char *module);

  //! Parse arguments from list of properties as it was arguments 
  //! of one module. Any pair <var,val> (resulting from a "var = val;" line 
  //! if the properties have been loaded from a file) are understood as
  //! a "--var val" on the command line.   
  //! @param module module to which adding arguments
  //! @param prop list of properties
  void parse(const char * module, ACOM::Properties prop);

  //! Parse arguments from list of properties as it was arguments 
  //! of one module. Any pair <var,val> (resulting from a "var = val;" line 
  //! if the properties have been loaded from a file) where var begins with 
  //! prefix is understood as  a "--var val" on the command line. Variables not
  //! begining with prefix are ignored.
  //! @param module module to which adding arguments
  //! @param prop list of properties
  //! @param prefix prefix use for recognition
  void parse(const char * module, ACOM::Properties prop, char * prefix);

  //! reset arguments
  void reset(void); 

  //! returns the list of modules
  std::list<std::string> modules(void) {
    return _modlist;
  }
  
  //! get args info of a module
  ArgVect& operator[]( const std::string& key);

  const ArgVect& operator[]( const std::string& key) const;

  //@name Iterator over the couple <key,value>
  //@{
  typedef std::map<std::string,ArgVect>::iterator iterator;
  typedef std::map<std::string,ArgVect>::const_iterator const_iterator;
    ///
  iterator begin() { return _map.begin(); }
    ///
  iterator end() { return _map.end(); }
    ///
  const_iterator begin() const { return _map.begin(); }
    ///
  const_iterator end() const { return _map.end(); }
  //@}
   const char * progname() { return _progname.c_str(); }
 protected:  
  std::map<std::string,ArgVect> _map;
  std::list<std::string> _modlist;
  std::string _progname;
};


} // namespace

#endif
