// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_netbase.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Jacques Briat, Thierry Gautier, Cyrille Martin
// ==========================================================================
#ifndef _ACOM_NETBASE_H_
#define _ACOM_NETBASE_H_

#include <inuk/acom_netconfig.h>
#include <inuk/acom_types.h>
#include <inuk/acom_comerror.h>
#include <inuk/acom_properties.h>
#include <string>
#include <list>
#include <map>
#include <inuk/pth>

//@Include: acom_netconfig.h acom_comerror.h acom_types.h

// ACOM namespace
namespace ACOM_NET_NAMESPACE {


/* fwd declaration */
class Network;

//! @defgroup NetDev Network Developers
//! @brief The 'Network Developers' group gathers classes intended to 
//! developers of networks
  
//! @defgroup NetUsers Network Users
//! @brief The 'Network Users' group gathers types intedended to 

//! @defgroup Request Request Types
//! @brief Request types
//! @ingroup NetDev

// ---------------------------------------------------------------------------
// NetworkBase class
//! @brief The virtual API for new Inuktitut interface Network.
//! @ingroup NetDev
//! @ingroup NetUsers
//! See network interface : it is split in serveral sub interfaces :
//! - NetworkBase   : properties, name, etc
//! - NetworkClient : what necessary to send request
//! - NetworkServer : what necessary to recv message
//! - NetworkCom    : get route, barrier, broadcast...
//! - NetworkAdmin  : what necessary add node
// ---------------------------------------------------------------------------

class NetworkBase {
public:
  NetworkBase();

  virtual ~NetworkBase();

  typedef uint16_t     LocalNodeNo; //!< type for node number

  //! kind of protocol
  enum ProtocolNo { 
    P_VOID =0,  //!< void message type
    P_AM   =1,  //!< Active Message: IOVector allocated & freed by the network
    P_WS   =2,  //!< Write & Signal: IOVector allocated & freed by application
    P_AWS  =3,  //!< Alloc, Write & Signal: IOVector allocated by the network, freed by app
    P_FREE      //!< index of the next free ProtocolNo
  };


  // --------------------------------------------------------------------------
  //! @brief Network Header for message
  //! 
  //!   This is the fixed part first sent on a channel
  //! @ingroup NetUsers
  //! @ingroup NetDev
  // --------------------------------------------------------------------------
  struct Header {
    uint32_t     size_eh;     //!< size of the extended header
    uint32_t     size_h;      //!< size of the service header (immediate data)
    uint32_t     size_d;      //!< size of the deferred data
    LocalNodeNo  dst;         //!< destination node
    LocalNodeNo  src;         //!< source node
    uint8_t      protocol;    //!< type of the protocol 
    uint32_t     id;          //!< service identifier
    uint32_t     iovsz_d;     //!< number of IOVectEntry of deferred data
    uint32_t     no;          //!< logical time stamp

    Header()                  //!< Constructor for Header class
     : size_eh(0), size_h(0), size_d(0), protocol(P_VOID), id(0), iovsz_d(0)
    {}
  };

  // -------------------------------------------------------------------------
  //! @brief State of a request
  //! @ingroup Request
  // -------------------------------------------------------------------------

  enum State {
      S_VOID =0,
      S_ALLOCATED,  //!< the request is allocated and can be initialized
      S_INITIALIZE, //!< the request is initialized and can be posted
      S_POSTED,     //!< the request has been posted into the network
      S_COMPLETED,  //!< the request is completed (data sent or recv), callback could be called
      S_CANCELLED,  //!< the request has been cancelled
      S_FREE,       //!< the request is freed, could be reused
      S_DESTROY     //!< the request will be destroyed
  };

  // -------------------------------------------------------------------------
  // BaseRequest class
  //! @brief Base class of request for the network
  //! @ingroup Request
  //!
  //!  This class serves as  base of the Request and Receipt classes
  // -------------------------------------------------------------------------
  class BaseRequest {
  public :
    /** default constructor
    */
    //!< default constructor
    BaseRequest(); 

    //!< Destructor
    virtual ~BaseRequest();                

    //!< Initializer of a request
    void init( Network* manager=0 );       

    //!< Get last error for request
    ComFailure::Code get_error();          

    //!< Set error code
    void set_error(ComFailure::Code e);    

    //!< Get request state
    State  get_state();
    
    //!< Set request state
    void set_state(State );

    //! Return a pointer to the network header 
    const NetworkBase::Header* get_header() const; 
    
#if 0 /* DEPRECATED */
    /** allocate the headed header : opaque bytes dependent of
        the implementation of a protocol on the network.
    */
    void* allocate_extended_header( size_t )
      throw (BadAlloc);
#endif

    //! set the headed header as a block of network allocate bytes 
    //!    : opaque bytes dependent of the implementation of a protocol 
    //!    on the network.
    void set_extended_header( void* b, size_t s);

    const Network* get_net() const; //!< Return a pointer to the Network 
    Network* get_net();

  public:
    Network*            _net;        //!< the network that manage the request
    State               _state;      //!< the state of the request
    ComFailure::Code    _error_no;   //!< error number
    NetworkBase::Header _header;     //!< header to pack & send or to recv & unpack 
    void*               _ext_header; //!< extended header specialized to network
    size_t              _size_eh;    //!< size sz of extended header

    friend class Network;
    friend class OutChannel;
  };


  // ---------------------------------------------------------------------------
  // Network interface
  // ---------------------------------------------------------------------------
  
  //! @name Network initialilization and termination
  //@{
  //!
  //! First stage initialization of a primary network
  //! @param argc pointer to argument count
  //! @param argv pointer to argument array
  //! @return 0 if successfull
  //! @remark All recognized arguments are deleted from the command line
  //! @warning On return the network is not necessarily initilized. 
  //! @warning A call to commit must be performed to terminate the initilialization
  //! This function is called to initialize a primary network.
  //! It is automatically called at initilization of the network layer for
  //! each primary network.  At the end of the first stage of initialisation,
  //! all local initializations are done, but communication cannot occurs yet
  virtual int initialize( int* argc, char** argv[] )
    throw (NetworkError) =0;

  //! First stage initialization of a secondary network
  //! @param argc pointer to argument count
  //! @param argv pointer to argument array
  //! @param netInit primary network to used to initialize the network
  //! @return 0 if successfull
  //! @remark All recognized arguments are deleted from the command line
  //! This function is called should to initialize a secondary network.
  //! It is automatically called at initilization of the network layer for
  //! each secondary network
  virtual int initialize( int* argc, char** argv[], acom::Network* netInit )
    throw (NetworkError) =0;

  //! Second stage of the initialisation process.
  //! At the end of the second stage of initialisation communication can occurs
  //! or during the call of initcommit in an other module (??). 
  //! This function is automatically called at initialization of the network 
  //! layer for each network
  //! @warning Call to this function must be done after a call to init
  //! @return 0 if successfull
  virtual int commit () throw (NetworkError) =0;

  //! Terminate a network
  //! This function is automatically called at termination of the network 
  //! layer for each network
  virtual int terminate () throw (NetworkError) =0;

  //@}


  // ----------------------------------------------------------------------
  //! @name Memory management
  // ----------------------------------------------------------------------
  //@{
  //! Allocate an immediate (??) buffer of sz bytes in network memory
  //! @param  sz (IN) size of the memory region in bytes 
  virtual void* allocate_data( size_t sz ) 
       throw (BadAlloc) =0;

  //! Deallocate a previously allocated immediate (??) buffer in network memory
  //! @param  data (IN) memory region to deallocate
  virtual void deallocate_data(  void* data ) 
       throw (InvalidArgument) =0;

  //! Add to  network memory a part of user memory
  //! @param  begin (IN)IN) beginning address of the memory block 
  //! @param  sz sz   (IN)in bytes of the memory block 
  //! @warning The zone is mapped. The really mapped zone is the rounding of 
  //! the zone in system page size unit. At this level, no registering of 
  //! concurrent mappings or overlapping of zones are done.
  virtual void  map_zone( void * begin, size_t sz) throw (BadAlloc);

  //! Retract from network memory a part of user memory 
  //! @param  begin (IN) beginning address of the memory block 
  //! @param  sz   (IN) in bytes of the memory block 
  //! @warning The zone is unmapped. At this level, no control of overlapping 
  //! of mapping or multiples mapping of the same zone are done..   
  virtual void  umap_zone( void * begin, size_t sz) throw (InvalidArgument);

  //@}

  // ----------------------------------------------------------------------
  //! @name Properties and name management
  // ----------------------------------------------------------------------
  //@{

  //! Get the name of the network. 
  const std::string& get_name( ) const;

  //! Set the name of the network. 
  const std::string& set_name(const std::string& name );

  //! Get the information attribute for the network.
  const std::string& get_property( const std::string& key ) const;

  //! Set the information attribute for the network.
  void set_property( const std::string& key, const std::string& value );

  //! Return all the properties for the network.
  const Properties& get_properties( ) const;
  
  //@}


  // -----------------------------------------------------------------------
  //! @name Rank and node counting
  // -----------------------------------------------------------------------
  //@{

  //! Return the number of node connected to this Network
  size_t node_count() const;
  //! Return the number of node connected to this Network
  size_t get_nodecount() const;

  //! set the identifier of the current node in the network
  void set_rank(LocalNodeNo n);

  //! Return the identifier of the current node in the Network
  LocalNodeNo rank() const;
  //! Return the identifier of the current node in the Network
  LocalNodeNo get_rank() const;

  //@}

  //! Return the identifier of the Network
  uint16_t get_id() const;

protected:
  uint16_t            _nid;          //!< id of the network instance
  Properties          _properties;   //!< properties for a network
  int                 _node_count;   //!< number of node in this network
  LocalNodeNo         _node_rank;    //!< rank of this node in this network
}; // -- end class network


/*
 * inline definition
 *
 */
// ---------------------------------------------------------------------------
// BaseRequest class
// --------------------------------------------------------------------------- 
inline NetworkBase::BaseRequest::BaseRequest()
  : _net(0), _state( S_VOID ), _error_no(ComFailure::OK), 
    _header(), _ext_header(0), _size_eh(0)
{ _state = S_ALLOCATED; }

inline NetworkBase::BaseRequest::~BaseRequest() 
{}

inline void NetworkBase::BaseRequest::init( Network* manager )
{ 
  _error_no = ComFailure::OK;
  _net = manager;
  _state = S_ALLOCATED; 
}

inline ComFailure::Code NetworkBase::BaseRequest::get_error() 
{ return _error_no; }

inline void NetworkBase::BaseRequest::set_error(ComFailure::Code e) 
{ _error_no = e; }

inline NetworkBase::State NetworkBase::BaseRequest::get_state() 
{ return _state; }

inline void NetworkBase::BaseRequest::set_state(State s)
{ _state = s; }

inline const NetworkBase::Header* NetworkBase::BaseRequest::get_header() const
{ return &_header; }

inline const Network* NetworkBase::BaseRequest::get_net() const 
{ return _net; }

inline Network* NetworkBase::BaseRequest::get_net()
{ return _net; }

inline void NetworkBase::BaseRequest::set_extended_header( void* b, size_t s)
{
  _ext_header = b;
  _size_eh = s;
}  


// ---------------------------------------------------------------------------
// NetworkBase class
// --------------------------------------------------------------------------- 
inline size_t NetworkBase::node_count() const 
{ return _node_count; }

inline size_t NetworkBase::get_nodecount() const 
{ return _node_count; }

inline void NetworkBase::set_rank(LocalNodeNo n)
{ _node_rank=n; }

inline NetworkBase::LocalNodeNo NetworkBase::rank() const 
{ return _node_rank; } 

inline NetworkBase::LocalNodeNo NetworkBase::get_rank() const 
{ return _node_rank; }

inline const Properties& NetworkBase::get_properties( ) const
{ return _properties; }

inline const std::string& NetworkBase::get_property( const std::string& key ) const
{ return _properties[key]; }

inline void NetworkBase::set_property( const std::string& key, const std::string& value )
{ _properties[key] = value; }

inline uint16_t NetworkBase::get_id() const
{ return _nid; }

} // end namespace ATH...
#endif 
