// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_netserver.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Jacques Briat, Thierry Gautier, Cyrille Martin
// ==========================================================================
#ifndef _ACOM_NETSERVER_H_
#define _ACOM_NETSERVER_H_

#include <inuk/acom_netbase.h>
#include <inuk/acom_netclient.h>

//@Include: acom_netconfig.h acom_comerror.h acom_types.h

// ACOM namespace
namespace ACOM_NET_NAMESPACE {

// ---------------------------------------------------------------------------
// NetworkServer class
// ---------------------------------------------------------------------------
/**@name NetworkServer
The virtual API for new Inuktitut interface Network.
The Network interface is splitted in serveral sub interfaces.
NetworkServer deals with receiving message.
*/
class NetworkServer {
public:
  /** */
  NetworkServer();

  /** */
  virtual ~NetworkServer();
  
  // ---------------------------------------------------------------------------
  // Receipt class
  // --------------------------------------------------------------------------- 
  /**@name Receipt
    Format of received request for the Network.  This class is the server side 
    descriptor of a active message request. It registers the immediate data 
    to received and maintains the state of progress of the receiving actions.
  */
 class Receipt : public NetworkBase::BaseRequest {
  public:
    /**  constructor
    */
    Receipt();

    /** destructor
    */
    ~Receipt();
    
    /** unpack immediate value ( return the size of unpacked data )
    */
    size_t unpack( void* d, size_t dsz);

    /** end of immediate values ?
    */
    bool eos( );

    /** Return the iovect entry
    */
    void get_iovect( IOVectEntry*& vector, IOVectSize& count);

    /** initializer of receipt ( allocator for iovector & buffers) from
      a byte sequence (immediate data) and an iovector (for defered data)
    */ 
    bool init(ByteSeq * bs , IOVectEntry * iov, size_t sz, Network* manager);

    /** initializer of receipt( allocator for iovector & buffers) from
      a iovect sequence (immediate data) and an iovector (for defered data)
    */
    bool init(IOVectSeq * iovs , IOVectEntry * iov, size_t sz, Network* manager);

    /** finalizer of a receipt; free all resources
    */
    void finish();

    /** Accessor to internal data
    */
    ByteSeqReader* get_datastream();
    
    /** Reset Receipt for data and iov streams 
    */
    IOVectSeq* get_packetbyte_vectseq();

  protected:
    IOVectSeq*    _iovs;
    ByteSeqReader _datastream;   // immediate  data
    IOVectEntry*  _iovbuff;      // vector of deferred data blocks
    size_t        _iovbuffsz;    // iovector size

    friend class Network;
  };


  // ---------------------------------------------------------------------------
  // CallContext struct
  // ---------------------------------------------------------------------------    
  /**@name Serveur part data structure that is used to make a call to a service
  */
  class CallContext : public PTH::Linkable<CallContext> {
  public:
    enum { ACOM_STACK_ALIGN = 8 };
  public:
    enum State {
      S_ALLOCATED,  /* allocated and initialized object */
      S_RESET,      /* cc is ready for receiving */
      S_EXECUTING,  /* a service call is under way */
      S_SUSPEND,    /* the CC was suspended and waiting for an event */
      S_QUEUED,     /* the CC queued */
      S_DEALLOCATED /* destroy and deallocated object */
    };

    /* public data field
    */
    Network*             network;    /// pointer to the network
    ComFailure::Code     error_no;   /// error number
    Receipt*             receipt;    /// receipt for data
    NetworkBase::Header* header;     /// header
    void*                ext_header; /// pointer to the extended header
    IOVectEntry*         iov;        /// pointer to the first io vector entry
    IOVectSize           iov_sz;     /// size of the io vector
    void*                cc;         /// pointer to user space a call context
    
  
    /* Constructor
    */
    CallContext();
    
    /* Destructor
    */
    virtual ~CallContext();

    /* Initialization
    */
    void init(Network* net, size_t capacity );

    /* reserve more memory for the stack
    */
    void reserve(size_t sz );
    
    /* Destruction
    */
    void destroy();

    /* Return a pointer into the stack to store sz bytes
    */
    void* allocate( size_t sz );

    /* Return the top of the stack
    */
    void* top();

    /* Increment the top of sz bytes and return new top
    */
    void* push( size_t sz );

    /* Return the current size of the stack
    */
    size_t size();

    /* Return the free size of the stack
    */
    size_t free_size();

    /* Return the offset from the base of an address 'a' into the stack
    */
    size_t get_offset(void* a);
    
    /* Return the address of base + offset
    */
    void* get_address(size_t o);

    /* Stack reset
       destroy _im_data_seq and finish the receipt
    */
    virtual void reset();
    
    /* acquire a reference on the call context: increment internal refcount
       by rc.
    */
    void acquire(int rc =1);

    /* is_acquired return true if the call context is acquired
       and could not be reused or released.
    */
    bool is_acquired();

    /* release a reference on the call context: short cut for
       network->release_callcontext( this ).
    */
    void release();

  public:
    /* private data fields */    
    PTH::Mutex    _access;
    volatile long _refcount;
    IOVectSeq     _im_data_seq;
    size_t        _stack_size;
    void*         _base;
    void*         _top;
    friend class NetworkServer;
  };

  // ---------------------------------------------------------------------------
  // NetworkServer interface
  // ---------------------------------------------------------------------------
  
  /** Allocate a new call context to treats incomming message
      The method MUST no has a default implementation.
      @param stack_size if the size of the required stack capability
  */
  virtual CallContext* allocate_callcontext( size_t stack_size = (size_t)-1) throw (BadAlloc) =0;

  /** Deallocate a call context
      The method MUST no has a default implementation.
  */
  virtual void deallocate_callcontext(CallContext*) throw (InvalidArgument) =0;
  
  /** resize the stack of the call context with capacity sz.
      Note that resizing will possibility return object is a new address.
      All variable the points into the stack of the call context should be
      saved and restore using get_offset/get_address on the callcontext object.
      @param ctxt the call context to be resized
      @param sz the request new size for the stack
      @retval the new call context
  */
  virtual CallContext* resize_callcontext(CallContext* ctxt, size_t sz ) throw (BadAlloc) =0;
    
  /** Acquire a call context for a new incomming communication.
      The network store a free list of callcontext : acquire_callcontext first 
      tries to allocate in the free list before call allocate_callcontext.
      The returned call context is acquired.
      This method is not thread aware with respect to the network object : 
      the caller should ensure that no multiple thread call it at the same time.
  */
  virtual CallContext* acquire_callcontext() throw (BadAlloc);
  
  /** Release a call context.
      The method releases the call context in the free list of call context
      iff no more reference has been acquired on the call context, i.e.
      iff the number of calls to CallContext::release is equals to the
      number of calls to CallContext::acquire.
      This method is not thread aware with respect to the network object :
      the caller should ensure that no multiple thread call it at the same time.
  */
  virtual void release_callcontext(CallContext* ) throw (InvalidArgument);
  
protected:
  /* free list of cc */
  virtual void clear();
  size_t                 _szlistcc;
  PTH::StackList<CallContext> _listcc;
}; // -- end class NetworkServer


/*
 * inline definition
 *
 */
// ---------------------------------------------------------------------------
// Receipt class
// --------------------------------------------------------------------------- 
inline NetworkServer::Receipt::Receipt()
  : NetworkBase::BaseRequest(), _iovs(0), _datastream(), _iovbuff(0), _iovbuffsz(0)
{}

inline NetworkServer::Receipt::~Receipt() 
{}

inline size_t NetworkServer::Receipt::unpack( void* d, size_t dsz )
{ return _datastream.unpack( d,dsz  ); }

inline bool NetworkServer::Receipt::eos( )
{ return _datastream.eos(); }

inline void NetworkServer::Receipt::get_iovect( IOVectEntry*& vector, IOVectSize& count) 
{ vector= _iovbuff; count= _header.iovsz_d; }

inline bool NetworkServer::Receipt::init(ByteSeq * bs , IOVectEntry * iov, size_t sz, Network* manager)
{ 
  NetworkBase::BaseRequest::init(manager);
  _datastream.init(bs); _iovbuff = iov; _iovbuffsz = sz;
  return true;
}

inline bool NetworkServer::Receipt::init(IOVectSeq * iovs , IOVectEntry * iov, size_t sz, Network* manager)
{ 
  NetworkBase::BaseRequest::init(manager);
  _iovs = iovs;
  _datastream.init(iovs); _iovbuff = iov; _iovbuffsz = sz;
  return true;
}

inline void NetworkServer::Receipt::finish() 
{
  _state = NetworkBase::S_DESTROY; _error_no = ComFailure::OK;
  _datastream.destroy(); 
  _iovbuff=0; _iovbuffsz = 0;
}

inline ByteSeqReader* NetworkServer::Receipt::get_datastream() 
{ return &_datastream; }

inline IOVectSeq* NetworkServer::Receipt::get_packetbyte_vectseq()
{ return _iovs; }



// ---------------------------------------------------------------------------
// CallContext class
// --------------------------------------------------------------------------- 
inline NetworkServer::CallContext::CallContext()
 : network(0), header(0), ext_header(0), iov(0), iov_sz(0), cc(0),
   _refcount(1), _stack_size(0), _base(0), _top(0)
{
  _access.init();
}

inline void* NetworkServer::CallContext::allocate(size_t sz )
{
  ACOM_ASSERT_DEBUG( ((char*)_top-(char*)_base) + sz < _stack_size, "call stack out of bound");
  if (sz ==0) return 0;
  void* retval = _top;
  // WARN tg: alignement here ?
  sz = (sz + ACOM_STACK_ALIGN-1) & ~(ACOM_STACK_ALIGN-1);
  _top = sz + (char*)_top;
  return retval;
}

inline void* NetworkServer::CallContext::top()
{
  return _top;
}

inline void* NetworkServer::CallContext::push(size_t sz )
{
  ACOM_ASSERT_DEBUG( ((char*)_top-(char*)_base) + sz < _stack_size, "call stack out of bound");
  if (sz ==0) return 0;
  _top = sz + (char*)_top;
  return _top;
}

inline size_t NetworkServer::CallContext::size()
{
  return ((char*)_top-(char*)_base);
}

inline size_t NetworkServer::CallContext::free_size()
{
  return _stack_size - ((char*)_top-(char*)_base);
}

inline size_t NetworkServer::CallContext::get_offset(void* a)
{
  return ((char*)a-(char*)_base);
}

inline void* NetworkServer::CallContext::get_address(size_t o)
{
  return (o+(char*)_base);
}

inline bool NetworkServer::CallContext::is_acquired()
{
  PTH::CriticalGuard critical( _access );
  return _refcount >1;
}

} // end namespace ATH...

#endif 
