// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_service.h,v $
// Copyright(c)'2001 by Apache Team
// acom_service.h : 
// Authors: Jacques Briat, Thierry Gautier
//
//
// ==========================================================================
#ifndef _ACOM_SERVICE_H_
#define _ACOM_SERVICE_H_

#include <inuk/acom_netconfig.h>
#include <inuk/acom_types.h>
#include <inuk/acom_iovect.h>
#include <inuk/acom_comerror.h>
#include <inuk/acom_network.h>
#include <map> 
#include <list> 

//@Include: acom_config.h acom_comerror.h 
namespace ACOM_NET_NAMESPACE {

class Service;

// --------------------------------------------------------------------------
//      Service Manager for active message and write&signal
// --------------------------------------------------------------------------
class ServiceManager {
public:
  /** Default cstor
  */
  ServiceManager();
  
  /** identifier of service
  */
  typedef  uint32_t  Id;

  /** Reserved entry for library namespace identifier
  */
  enum ReservedLNS {
     SYSTEM_LNS  = 0,
     TAKTUK_LNS  = 1,
     CORBA_LNS   = 2,
     TCP_LNS     = 3,
     DNS_LNS     = 4,
		 NETDRAW_LNS = 5,
		 SPREAD_LNS	 = 6,
     SOCKNET_LNS = 7,
     A1_LNS      = 8,
     ICS_LNS     = 9,
     DEFAULT_LNS = 10 /* default library namespace for user defined service */
  };

  /** Allocate a new free LNS 
      @return a new valid LNS 
      @exception OutOfRange if no more LNS could be allocated
   */
  int allocate_lns() throw (OutOfRange);

  /** Deallocate a LNS 
      @param  a LNS 
      @exception InvalidArgument if lns is a bad value
   */
  void deallocate_lns( int lns ) throw (InvalidArgument);


  /** Bind a service stub object to an identifier
    The lifetime of the object obj should be greather than the 
    usage of the service itself.
    @param lns  the (integer) namespace for the library
    @param sid  the (integer) identifier for the service
    @param obj  the service stub object
    @return an error code
    @exception InvalidArgument
  */
  void bind( int lns, Id sid,  Service* obj ) throw(InvalidArgument);

  /** Bind a service stub object to an identifier
    Bind the service identifier in the DEFAULT_LID namespace.
    @param sid  the (integer) identifier for the service
    @param obj  the service stub object
    @return an error code
    @exception InvalidArgument
  */
  void bind( Id sid,  Service* obj ) throw(InvalidArgument)
  { bind( DEFAULT_LNS, sid, obj); }


  /** Resolve a service stub object using its identifier
    If the identifier is not found, the stub of a null service 
    (that do nothing) is returned.
    @param id  the (integer) identifier
    @return the service stub object
    @exception InvalidArgument if invalid id
  */
  Service* resolv( Id id) throw(InvalidArgument);

private:  
  std::map<Id, Service*>  _services;
  int                     _nextlns;
  std::list<int>          _freelns;
};


// --------------------------------------------------------------------------
//      Service Base for all kinds of services
// --------------------------------------------------------------------------
/** Service.
  A service is an object that implement one of the following call methods
  (the method implemented in derived class depend on the way to use the
  service).
  The user that want to implement a service should create a class that
  derivated from Service and re-implement a method 'call'.
  Before any use, the service should be registered to the manager. This
  can be done by making a call to Service::Manager.bind( id, serv) or
  rebind on the static object Manager. See ServiceManager class for
  a reference on methods on the manager.

  The user that want to implement a service with a specific header
  should define a structure of name Header. The field of that structure
  could be set up during ActiveMessage or WriteAndSignal protocol call.
@see ProtocolService
*/
class Service {
public:
  typedef ServiceManager::Id Id;

  /** Default CSTOR
  */
  Service();

  /** Default DSTOR
  */
  virtual ~Service();

  /// Void service header
  //struct Header {};

  /** Service information accessors
  */
  Service::Id get_id() const { return _id; }

  /** Entry point for a service.
    Execution of this method should not be a blocking call or not post
    a message to a network .
    This method is called for all protocols when all data of the message
    have been received. The CallContext is used in input mode.
    Default implementation is to make a call to the 'call' method.
    @param call_cc (IN) the call context.
  */ 
  virtual void server_call( Network::CallContext* call_cc ) throw(ComFailure);

  /** Write and signal address decoder.
    The field iov of the call context structure MUST be filled.
    If there is a mismatch between send blocks and received blocks, network behaviour is impredictible.
    The 'cc' field could be used to pass extra parameters to the call of the
    service.
    Default implementation is to make a call to the 'get_iovect' method.
    @param call_cc (IN) the call context.
  */
  virtual void server_get_iovect( Network::CallContext* call_cc ) throw (ComFailure);

  /** Entry point for a service.
    Execution of this method should not be a blocking call.
    This method is called on an active message remote method call
    or a write and signal method call.
    @param  error_no (IN)error code if any
    @param receipt (IN) received request descriptor from which unpack immediate data
    @param  header (IN) the header of the incomming request 
    @param vector (IN) the addresses of deferred data blocks
    @param count (IN) iovector size
    @param ctxt (IN) a pointer to opaque data structure return by get_iovect 

  */ 
  virtual void call( 
      ComFailure::Code error_no,
      Network::Receipt* receipt,
      Network::Header* header, 
      IOVectEntry* vector,
      IOVectSize count,
      void* ctxt
  ) throw(ComFailure) =0;


  /** Write and signal address decoder.
    If not defined, use "active message" allocation of the stack
    done by the network.
		@param receipt (IN) the handle to get immediate data
    @param  header (IN)the header of the incomming request from which unpack immediate data
    @param  vector  iovector (IN-OUT) whose entries MUST be filled.
    @param count (IN) size of the iovect  to filled with block address
    \warning  If there is a mismatch between send blocks and received blocks, network behaviour is impredictible.
    @retval a pointer to a value that will be pass as 'ctxt' in the call to 'call' 
  */
  virtual void* get_iovect( 
    Network::Receipt* receipt,
    Network::Header* header, 
    IOVectEntry* vector, IOVectSize count
  ) =0;

  /** The manager of all services
  */
  static ServiceManager Manager;

protected:
  friend class ServiceManager;
  void set_id( Service::Id id ) { _id = id; }
  Service::Id _id;
/*
  PTH::Mtx    _access;
*/
};


/** WSService type to force implementation of method depending
    of their calling protocol
*/
typedef Service ServiceWS;

/** AMService type to force implementation of method depending
    of their calling protocol
*/
class ServiceAM : public Service {
public:
  void* get_iovect(
    Network::Receipt* receipt,
    Network::Header* header,
    IOVectEntry* vector, IOVectSize count
  ) 
  { return 0; }
};

class ServiceCC : public Service {
protected:
  void call( 
      ComFailure::Code error_no,
      Network::Receipt* receipt,
      Network::Header* header, 
      IOVectEntry* vector,
      IOVectSize count,
      void* ctxt
  ) throw(ComFailure)
  { }

  void* get_iovect(
    Network::Receipt* receipt,
    Network::Header* header,
    IOVectEntry* vector, IOVectSize count
  ) 
  { return 0; }
};


/** AWSService type to force implementation of method depending
    of their calling protocol
*/
typedef ServiceAM ServiceAWS;


/*
 * inline definitions
 *
 */
inline void Service::server_call( Network::CallContext* c ) throw(ComFailure)
{
  this->call( c->error_no, c->receipt, c->header, c->iov, c->iov_sz, c->cc );
}

inline void Service::server_get_iovect( Network::CallContext* c ) throw (ComFailure)
{
  c->cc = this->get_iovect( c->receipt, c->header, c->iov, c->iov_sz );
}

} // - namespace ACOM_..


#endif /* _ACOM_SERVICE_H */
