// =========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_stream.h,v $
// (c) INRIA, projet APACHE, 2001
// Author: T. Gautier
//
//
// =========================================================================
#ifndef _ACOM_STREAM_H_
#define _ACOM_STREAM_H_

#ifndef WIN32
#include <inuk/acom_netconfig.h>
#endif

#include <inuk/acom_error.h>
#include <inuk/acom_iovect.h>

#include <vector>
#if 0
#include <hash_map.h>
#endif

namespace ACOM_NET_NAMESPACE {
class Format;

/*!  
   \brief Root class for packet of immediate values. 
This class provides the basic operators to pack/unpack data in an bounded byte 
buffer. 
*/
class BytePacket {
protected :
  void * _scan;  /// scan pointer
  size_t _sz;    /// remaining size to fill or to be unpacked 

  /** This method may be overloaded in derived class to substitute a new 
      filled buffer to the emptied one
  */
  virtual bool flop() { return false ;}

public :
  /** default constructor
  */
  BytePacket(): _scan(0), _sz(0) {}

  /** virtual destructor
  */
  virtual ~BytePacket() {}

  /** initiator : buffer size + buffer are given 
  */
  void init(void *b, size_t bsz) 
  { _scan =b; _sz =  bsz; }

  /** destroy  a buffer
  */
  virtual void  destroy() {}
  
  /** pack bytes
    @param ssz (IN) bytes count  to be packed
    @param  s (IN)  IN-> first byte to be packed 
    @retval the size of packed bytes 
  */
  size_t pack( const void *  s, size_t ssz) { 
    if (ssz<=_sz) {
      memcpy(_scan ,s, ssz);
      char* tmp = (char*)_scan;
      _scan= tmp + ssz;
        _sz-=ssz; 
    } else {
      memcpy(_scan , s, _sz);
      char* tmp = (char*)_scan;
      _scan = tmp+_sz;
      ssz=_sz; _sz=0; 
    }
    return ssz;
  }

  /** unpack bytes 
    If all bytes have been unpacked return 0, else return the number of bytes 
    really unpacked 
    @retval the size of unpacked bytes 
    @param dsz (IN) bytes count  to be unpacked
    @param  d (IN/)  IN-> first byte where to unpack
  */
  size_t unpack( void* d, size_t dsz)
  { 
	//	std::cout << getpid() << " BytePacket::unpack enter: _sz=" << _sz << std::endl;
    if (dsz<=_sz) {
      memcpy( d , _scan ,  dsz);
      char* tmp = (char*)_scan;
      _scan = tmp + dsz;
      _sz-=dsz;   // all read !
    } else {
      memcpy( d,  _scan,  _sz);
      char* tmp = (char*)_scan;
      _scan = tmp + _sz; 
      dsz=_sz;_sz=0;
    }
	//	std::cout << getpid() << " BytePacket::unpack exit: _sz=" << _sz << " return: " << dsz << std::endl;
    return dsz; 
  }

  /** check end of packet
  */
  virtual  bool eop(){ return ( _scan==0) || (_sz==0);}
};



/*! 
   \brief Allocator for  byte buffer (default) 
Such allocators are used to allocate or free fixed sized blocks. 
Allocatcation is speeded by using a free block cache. 
The size of blocks is fixed at allocator's creation time. It is thread unsafe.
*/
typedef PTH::SyncByteAllocator ByteCellAllocator;


/*!  
   \brief Root class for sequence of packed immediate values. 
This class provides the basic operators to pack/unpack data in an  byte sequence. 
Object of this class appears as a continuous sequence of bytes which size is 
dynamic. It is implemented as an iovector sequence ( IOVectSeq )of fixed size 
buffers dynamically allocated. 

When an overflow is detected, a virtual function (OutByteSeq::ping() or 
OutByteSeq::pong() is called to provide  a new buffer. These functions must 
invoke a buffer allocator matching memory placement constraints of given networks. 
This allocator must be given when a sequence is created.

This sequence supports only one access mode at a  time (read THEN write). 
Before read, ByteBuffer::packEos() must be issued. If not, effect of following 
reads are undefined. It is thread unsafe.  The  cumulated  number of bytes 
is maintained  for a sequence.
*/
class ByteSeq : public BytePacket, public IOVectSeq  {
protected :
  /**  block allocator
  */
  ByteCellAllocator* _bufA;
  void * _buf;   /// current buffer address
  size_t _bufsz; /// current filled buffer size
public :
  /** default constructor
  */
  ByteSeq()
   : BytePacket(), IOVectSeq(), _bufA(0), _buf(0), _bufsz(0) 
  {}

  /** initiator : 
      buffer size, iovector size and buffer allocator are given  ( true if OK )
  */
  bool init ( size_t bsz, ByteCellAllocator* bufa,
              int iovsz, IOVectCellAllocator * iova) 
  {  _bufA=bufa;_bufsz=bsz;_buf=_bufA->alloc();
      BytePacket::init(_buf,bsz );
    return (IOVectSeq::init(iovsz, iova) && _buf);
  } 

  /** virtual destructor
  */
  virtual ~ByteSeq() {}

  /** destroy a stream 
    A stream must be close after writed & before  send or after read)
  */
  void destroy() 
  { 
    IOVectEntry iov;
    if(!_bufA)return;
    IOVectSeq::IteratorEntry iE=beginEntry();
    while (iE.get(&iov))
      if (iov.data != _buf)
      {
        if(iov.data)    
          _bufA->free(iov.data);
      }
    if (_buf) _bufA->free(_buf);
    IOVectSeq::destroy();
    _bufA=0;_buf=0; _bufsz=0;
  }


  /**  Function to switch to next buffer when current is completely packed or unpacked.
    Gets the next buffer (if possible) for write.
    This method may be overloaded in derived class to substitute a new buffer 
    to the filled one
  */
  virtual bool flip();


  /** Procedures to build a packed stream are inherited from ByteBuffer
  */
  //@{ 
  /** push end of stream
  */
  void packEos()
  {
    int sz=_bufsz-_sz;
    if( _buf && (sz>0))IOVectSeq::put(_buf,sz ); 
    IOVectSeq::putEos();
  }

  /** pack bytes
    @param ssz (IN) bytes count  to be packed
    @param  s (IN)  IN-> first byte to be packed 
    @retval the size of packed bytes 
  */
  size_t pack( const void* s, size_t ssz) 
  { 
    size_t packed_sz = 0;
    if(_buf==0) return 0;
    for(;;) 
    { 
      packed_sz += BytePacket::pack(((char *) s + packed_sz), 
                                    ssz - packed_sz );
      if(packed_sz == ssz) return ssz;
      if(!flip()) return packed_sz;
      //      std::cout << "flip: " << packed_sz << "/" << ssz << std::endl;
    }
  }
  //@}
  

  /** Size of the byte sequence (Only defined after packing eos)
  */
  int getFillSize()
  {
    return IOVectSeq::getByteSize();
  }
};



class ByteSeqReader : public BytePacket {
protected :
  IOVectSeq::IteratorEntry _iE; /// ByteSeq iterator
  size_t _seq_sz; /// total Byte seq size 
public :
  /** default constructor
  */
  ByteSeqReader()
   : BytePacket(),_iE() ,_seq_sz(0)
  {}

  /** initiator : 
      built from a ByteSeq ( true if OK )
  */
  bool init (ByteSeq * bS ) 
  { _seq_sz=bS->getFillSize();
    _iE=bS->beginEntry(); 
     _iE.get(&_scan,&_sz) ;
     bool err = BytePacket::eop();
  // std::cout << " SCAN " << _scan << "  " << _sz<<" EOP " << err << std::endl ;
    return err;
  } 

  /** initiator : 
      built from a IOVectSeq ( true if OK )
  */
  bool init (IOVectSeq * iovS ) 
  { _seq_sz=iovS->getByteSize();
    _iE=iovS->beginEntry();  _iE.get(&_scan,&_sz) ;
     bool err = BytePacket::eop();
  // std::cout << " SCAN " << _scan << "  " << _sz<<" EOP " << err << std::endl ;
    return err;
  } 

  /** virtual destructor
  */
  virtual ~ByteSeqReader() { _seq_sz=0; }

  /** get the next bufffer (if any) for read
  */
  virtual bool flop() { 
     return _iE.get(&_scan,&_sz);
  }


  /** Procedures to read a packed stream are inherited from ByteBuffer
    */
  //@{ 
  /** unpack bytes 
    If all bytes have been unpacked return 0, else return the number of bytes 
    really unpacked 
    @retval the size of unpacked bytes 
    @param dsz (IN) bytes count  to be unpacked
    @param  d (IN/)  IN-> first byte where to unpack
  */
  size_t unpack( void *  d, size_t dsz) 
  { 
    size_t unpacked_sz = 0;
    if(_scan==0) return 0;
    for(;;) { 
      unpacked_sz += BytePacket::unpack(((char *) d + unpacked_sz), 
                                        dsz - unpacked_sz);
      if (unpacked_sz==dsz) return dsz;
      if (!flop()) return unpacked_sz;
      //      std::cout << "flop: " << unpacked_sz << "/" << dsz << std::endl; 
    }
  }
  //@}
  
  
  /** check end of stream
  */
  bool eos()
  {
    bool x,y;
    y=BytePacket::eop();
    x=_iE.eos();
	//    std::cout << " IOV status: EOS: " << x  << " EOP: " << y << std::endl;
    return  x && y;
  }

  /** Size of the byte sequence (Only defined after packing eos)
  */
  int getFillSize()
  {
    return _seq_sz;
  }
};


/*!  
   \brief This class Map defines a map of the memory before sending
   data. 
   Three kinds of mode of description are defined: immediate
   mode (IA), differed mode (DA) and differed cyclic mode (DAC). 
   All data of the first one will be copied into a local buffer 
   passed as argument of the construction.
   In case of DAC insertion a local lookup table for already
   inserted data.
   The construction of the Map have in parameter the buffer to
   store the IA kind of data  and the pointeur of an IOVector.
*/
class Map {
public:
  enum Mode { 
    IA,   /* immediate access of value */ 
    DA,   /* differed access of value */
    DAC   /* differed access of possibly cyclic pointer value */
  };
  enum {
    THRESHOLD_IA = 32 /* threshold to pack putted data */
  }; 
protected:
  /** Function called when the array of mode 'm' is fill.
    The return value is 'true' is a the stream process of defining
    a format could be continued, and 'false' is this process should
    be stop.
    @param m the mode of insertion into the stream that make overflow
    @param add_len the size requested at insertion
  */
  virtual bool overflow( Mode m, size_t add_len );

protected:
};



/*!  
   \brief  data stream ..................gestion du compromis emission de copie
(emballage) ou emission en place 

- A LIRE AVANT UTILISATION DES STREAMS
La semantique de la definition des streams pour la definition des formats
fonctionnent de la maniere suivante :
  - 1/ l'operateur << est appele lors de l'envoi effectif des donnees. 
Lors de l'appel a cet operateur, deux tableaux sont generes : les donnees
immediates (les types de base de C++) sont copiees dans un buffer
tandis que les donnees non immediates (cf exemple: les vecteurs de la STL)
sont stockes dans un io vecteur (pointeur sur le debut + taille en octets).
Ensuite un message est envoye (cf a1jit_gencom.cpp) avec dans la partie
header du message les donnees immediates. Au cours de la formation de
ce message aucune copie des iovecteurs n'est effectuee. Par exemple
lors de l'envoi d'un tableau (cf exemple vecteur ci-dessous) la taille
est copiee dans la partie header du message tandis que le contenu du
tableau ne l'est pas.
  - 2/ l'operateur >> est appele lors de la reception effective d'une donnee. 
Cet operateur est appelle sur un IStream dont les valeurs immediates
sont accesibles puisqu'arrivant dans la partie header du message. Les
donnees non immediates (differees) ne sont pas accessible et lors
de l'execution des instructions de lecture de ces donnees un io vecteur
est stocke dans le stream.
Cette operateur >> est appelle essentiellement pour generer un iovecteur
des donnees afin de les recevoir directement en memoire.
Lors de la reception des donnees en provenance du reseau, le systeme
de communication Inuktitut tire ces donnees directement dans leurs
emplacement memoire.

En conclusion: il n'est pas permit lors de la reception de vouloir
lire une donnee d'un tableau dans la definition de l'operateur >> :
par exemple dans l'extrait de code suivant:
\code
  template<class T>
  inline IStream& operator>> (IStream& m, std::vector<T>& A )
  {
    size_t sz;

    // - lecture d'un entier: valeur immediate = OK
    m >> sz;     

    A.resize(sz);
    m.read( IStream::DA, &A[0], sz*sizeof(T));

    // - lecture d'une valeur dans une donnee differee = NO OK
    if (A[i] == 3) { 
      ....
    }
    return m;
  }
\endcode
Le test (A[i] == 3) n'est pas possible et a un comportement indefini.
Ces operateurs << et >> doivent etre vu comme des operateurs de
definition de format, sans controle possible sur les valeurs lus autres
que les valeurs immediates.
*/
class OStream : public Map {
public:
  /** Cstor */
  OStream( ByteSeq* datastream, IOVectSeq* iovstream )
   : Map(), _current_id(0), _datastream(datastream), _iovstream(iovstream)
  { }

  /** Dstor */
  virtual ~OStream() {}
  
  /** Write */
  virtual void write( const Format* f, Mode m, const void* data, size_t count );

  /** Write immediate data */
  virtual void write( const Format* f, const void* data, size_t count );

protected:
  /* contains set of pointer already inserted into the stream 
    _map_cylic[ptr] = id -> the ptr is already inserted and id
    is the key for communication in place of the data.
  */
#if 0
  typedef std::hash_map<void*,int> LookupTable; 
  LookupTable _map_cyclic; 
#endif
  int        _current_id;
  ByteSeq*   _datastream; /* stream of immediate data */
  IOVectSeq* _iovstream;  /* stream of deferred data blocks */
};


class IStream : public Map {
public:
  /** Cstor */
  IStream( ByteSeqReader* datastream, IOVectEntry* iov, size_t iovsz )
   : Map(), _current_id(0), _datastream(datastream), _iov(iov), _iovsz(iovsz)
  {}

  /** Dstor */
  virtual ~IStream() {}
  
  /** Read  */
  virtual void read( const Format* f, Mode m, void* const data, size_t count );

  /** Read immediate data */
  virtual void read( const Format* f, void* const data, size_t count );

protected:
  /* contains set of pointer already inserted into the stream 
    _map_cylic[id] = ptr -> the key id has been already inserted 
    and ptr is the address of the memory that should contains the data.
  */
#if 0
  typedef std::hash_map<int,void*> LookupTable; 
  LookupTable _map_cyclic; 
#endif
  int            _current_id;
  ByteSeqReader* _datastream;   // immediate  data
  IOVectEntry*   _iov;          // vector of deferred data blocks
  size_t         _iovsz;
};


/*
 * inline definition
 */

 
} // namespace ACOM_NET_NAMESPACE

#endif
