// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_atomic.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Authors: J. Briat &  T. Gautier
// $Id: pth_atomic.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================

#ifndef _PTH_ATOMIC_H_
#define _PTH_ATOMIC_H_

/*!
\file  
\brief Atomic counter for needs atomic counter for reference counting,
threads counting , synchronization etc... 
\author  J. Briat T. Gautier 
\date 2001

 Inuktitut0 INRIA.

\sa PTH::AtomicCounter
*/

#include <inuk/pth_ns.h>
#if !defined( PTH_USE_ARCH_X86  ) 
/* based on using NTMutex */
#include <inuk/pth_pthreadsync_notrace.h>
#endif


namespace PTH {

// -----
// Atomic Counter
// -----
/** AtomicCounter
This class implements integer counter with atomic in(de)crement 
*/


class AtomicCounter
{
private:
#if defined( PTH_USE_ARCH_X86  ) 
   /// counter value   ( based native X86 atomic operations  \sa atomic_t 
    volatile int _value; 
#else
    /// counter value
    volatile int _value; 
    /// lock for exclusive access
    PTH::NTMutex _atomicmutex;
#endif
      
public :
    /// destructor.
    virtual ~AtomicCounter();
    
    /// initialize the counter. 
    AtomicCounter(int x=0);

    
                
                /// increment  the counter by some value  
    void add(int v);
    /// increment  the counter by some value   
    int operator+=(int v);
                
    /// decrement   the counter by some value Result is the new value
    int sub(int v) ;

    // decrement   the counter by some value Result is the new value
    int operator-=(int v);
               
    /// increment by one the counter.
    void operator++();

    /// decrement by one the counter. Result is the new value
    int operator--();

    // test and inc  true if result is 0 

    // test and dec  true if result is 0 

    // test and add  true if result is <0 

    /// clear a mask
    void clearMask(int v);
    /// reset a mask
    void setMask(int v);
                 
    /// Read the  value
    int getValue()const {return _value;};
    /// Read the  value
    operator int() const {return _value;};
                
    /// Set the  value
    void setValue(int v){ _value=v; }
};


#if defined(PTH_USE_ARCH_X86) 
// code extracted from linux kernel code include/asm/atomic.h */

#if defined( PTH_USE_ARCH_SMP)
#  define LOCK "lock ; "
#elif defined( PTH_USE_ARCH_MONO)
#  define LOCK ""
#endif   
 /**
 * used only for implementind AtomicCounter
 *
 * This struct is to be sure gcc doesn't try to be clever and move things around
 * on us. We need to use _exactly_ the address the user gave us,
 * not some alias that contains the same information.
 */

typedef struct { 
     volatile int counter;/**< counter value */
 } atomic_t;






/**
 * atomic_add - add integer to atomic variable
 * @param i: integer value to add
 * @param v: pointer of type atomic_t
 * 
 * Atomically adds @param i to @param v.  Note that the guaranteed useful range
 * of an atomic_t is only 24 bits.
 */

static __inline__ void atomic_add(int i, atomic_t * v)
{
  __asm__ __volatile__(
    LOCK "addl %1,%0"
    :"=m" (v->counter)
    :"ir" (i), "m" (v->counter));
}


/**
 * atomic_sub - subtract the atomic variable
 * @param i: integer value to subtract
 * @param v: pointer of type atomic_t
 * 
 * Atomically subtracts @param i from @param v.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */

static __inline__ void atomic_sub(int i, atomic_t * v)
{
  __asm__ __volatile__(
    LOCK "subl %1,%0"
    :"=m" (v->counter)
    :"ir" (i), "m" (v->counter));
}

/**
 * atomic_sub_and_test - subtract value from variable and test result
 * @param i: integer value to subtract
 * @param v: pointer of type atomic_t
 * 
 * Atomically subtracts @param i from @param v and returns
 * true if the result is zero, or false for all
 * other cases.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */

static __inline__ int atomic_sub_and_test(int i, atomic_t * v)
{
  unsigned char c;

  __asm__ __volatile__(
    LOCK "subl %2,%0; sete %1"
    :"=m" (v->counter), "=qm" (c)
    :"ir" (i), "m" (v->counter) : "memory");
  return c;
}

/**
 * atomic_inc - increment atomic variable
 * @param v: pointer of type atomic_t
 * 
 * Atomically increments @param v by 1.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */ 

static __inline__ void atomic_inc(atomic_t * v)
{
  __asm__ __volatile__(
    LOCK "incl %0"
    :"=m" (v->counter)
    :"m" (v->counter));
}


/**
 * atomic_dec - decrement atomic variable
 * @param v: pointer of type atomic_t
 * 
 * Atomically decrements @param v by 1.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */ 

static __inline__ void atomic_dec(atomic_t * v)
{
  __asm__ __volatile__(
    LOCK "decl %0"
    :"=m" (v->counter)
    :"m" (v->counter));
}


/**
 * atomic_dec_and_test - decrement and test
 * @param v: pointer of type atomic_t
 * 
 * Atomically decrements @param v by 1 and
 * returns true if the result is 0, or false for all other
 * cases.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */ 

static __inline__ int atomic_dec_and_test(atomic_t * v)
{
  unsigned char c;

  __asm__ __volatile__(
    LOCK "decl %0; sete %1"
    :"=m" (v->counter), "=qm" (c)
    :"m" (v->counter) : "memory");
  return c != 0;
}


/**
 * atomic_inc_and_test - increment and test 
 * @param v: pointer of type atomic_t
 * 
 * Atomically increments @param v by 1
 * and returns true if the result is zero, or false for all
 * other cases.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */ 

static __inline__ int atomic_inc_and_test(atomic_t * v)
{
  unsigned char c;

  __asm__ __volatile__(
    LOCK "incl %0; sete %1"
    :"=m" (v->counter), "=qm" (c)
    :"m" (v->counter) : "memory");
  return c != 0;
}


/**
 * atomic_add_negative - add and test if negative
 * @param v: pointer of type atomic_t
 * @param i: integer value to add
 * 
 * Atomically adds @param i to @param v and returns true
 * if the result is negative, or false when
 * result is greater than or equal to zero.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */ 

static __inline__ int atomic_add_negative(int i, atomic_t * v)
{
  unsigned char c;

  __asm__ __volatile__(
    LOCK "addl %2,%0; sets %1"
    :"=m" (v->counter), "=qm" (c)
    :"ir" (i), "m" (v->counter) : "memory");
  return c;
}

/* These are x86-specific, used by some header files */

static  __inline__ void  atomic_clear_mask(int mask,atomic_t * v) 
{
  __asm__ __volatile__(LOCK "andl %0,%1" 
  : : "r" (~(mask)),"m" (v->counter)  : "memory");
}



static  __inline__ void atomic_set_mask(int mask, atomic_t * v) 
{
__asm__ __volatile__(LOCK "orl %0,%1" 
: : "r" (mask),"m" (v->counter) : "memory");
}



    /// destructor.
inline AtomicCounter::~AtomicCounter(){};

    /// initialize the counter.  
inline AtomicCounter::AtomicCounter(int x):_value(x){};

    /// increment  the counter by some value   
inline void AtomicCounter::add(int v){ atomic_add(v,( atomic_t*)&_value);     }
inline  int AtomicCounter::operator+=(int v){add(v); return _value;};

    /// decrement   the counter by some value Result is the new value
inline int AtomicCounter::sub(int v) { atomic_sub(v,( atomic_t*)&_value)  ; return _value;   }
inline int AtomicCounter::operator-=(int v){sub(v); return _value;}

    /// increment by one the counter.
inline void AtomicCounter::operator++(){ atomic_inc(( atomic_t*)&_value);};

    /// decrement by one the counter. Result is the new value
inline int AtomicCounter::operator--(){ atomic_dec(( atomic_t*)&_value); return _value;};


    // clear a mask
inline void AtomicCounter::clearMask(int v){atomic_clear_mask(v,( atomic_t*)&_value  );}
    /// reset a mask
inline void AtomicCounter::setMask(int v)  {atomic_set_mask(v,( atomic_t*)&_value  );}
    

#else
//  Atomic counter in a Posix synchronization

    /// initialize the counter. 
inline AtomicCounter::AtomicCounter(int x)
 : _value(x), _atomicmutex()
{ _atomicmutex.init(); }

    /// destructor.
inline AtomicCounter::~AtomicCounter(){}

    /// increment by one the counter.
inline void AtomicCounter::operator++()
{ NTCriticalGuard sec(_atomicmutex); ++_value; }

    /// decrement by one the counter. Result is the new value
inline int AtomicCounter::operator--()
{ NTCriticalGuard sec(_atomicmutex); return --_value; }

    /// increment  the counter by some value     POSIX 
inline void AtomicCounter::add(int v)
{ NTCriticalGuard sec(_atomicmutex); _value+=v; } 
inline int AtomicCounter::operator+=(int v)
{ NTCriticalGuard sec(_atomicmutex); return _value+=v; }

    /// decrement   the counter by some value Result is the new value
inline int AtomicCounter::sub(int v) 
{ NTCriticalGuard sec(_atomicmutex);  _value-=v; return _value; }
inline int AtomicCounter::operator-=(int v)
{ NTCriticalGuard sec(_atomicmutex);  _value-=v; return _value;}
    /// set a mask
inline void AtomicCounter::clearMask(int v)
{ NTCriticalGuard sec(_atomicmutex);  _value&=~v; }
    /// reset a mask
inline void AtomicCounter::setMask(int v)
{ NTCriticalGuard sec(_atomicmutex);  _value|=v; }

#endif 

} // end namespace
#endif
