// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_basic.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// $Id: pth_basic.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================
#ifndef _PTH_BASIC_H
#define _PTH_BASIC_H

#include <inuk/pth_ns.h>
#include <inuk/pth_error.h>
#include <inuk/pth_macro.h>
#include <stdlib.h>
#include <iostream>


namespace PTH {
/*! 
\file  
\brief  linked list queue management (thread unsafe)
\author  J. Briat
\date 2001

linked list queue management (thread unsafe)
\sa JTH::Linkable JTH::StackList JTH::Queable JTH::Queue

and a simple cell allocator with free blocks caching
\sa PTH::CellAllocator PTH::BlockAllocator 
*/


// forward declaration of StackList to set  friends 
template< class _CELLTYPE_  > class StackList;
template< class _CELLTYPE_  > class List;
template < class _CELLTYPE_ > class Linkable;


//\ingroup Glib
/*! 
\brief this class is the  base class to make cells linkable in single linked list

To be (de)linkable in single linked list, a cell  must be instance of a class derived from class Linkable. A linkable cell may be in at most one queue.
   A  linkable class must be declared as follow :
\code 
   class X : public Linkable<X> {
    ..; members etc...
   };
\endcode 
*/
template < class _CELLTYPE_ >  
class Linkable {
protected: 
  _CELLTYPE_* _next; 
public :
  /// Constructor
  Linkable():_next(0){}; 

  /// Destructor. The cell must not in a queue
  ~Linkable()throw (PosixError){
    ASSERT( !_next , 
        PosixError(SUFX("[Linkable :] destroy an linked queue cell")));
  };

  /// get successor
  _CELLTYPE_* next() const throw()
  { return _next; }

  friend class  StackList<_CELLTYPE_>; 
  friend class  List<_CELLTYPE_>; 
};
 


// \ingroup Glib
 /*!  \brief Stack linked list   (thread unsafe)
  The  StackList descriptor is linked to the top element.
  We can define a stack list  of Linkable cells  :
  \code 
  StackList<X> Stk_of_X;
  \endcode  
  \warning In deletion, queue must be not empty. In insertion, cell
  must be defined ( not 0)
*/
template < class _CELLTYPE_ > 
class StackList{
  ///   top
  _CELLTYPE_* _top;            

public: 
  /// Constructor of an empty queue
  StackList()
   : _top(0) 
  {}

  /// Destructor. The queue must be  an empty queue
  ~StackList() throw (PosixError) {
    ASSERT( empty(), PosixError(SUFX("[StackList :] destroy a non empty list")));
  }

  /// check empty stack
  bool empty() const throw()
  { return _top ==0; } 

  ///  get  stack size
  size_t size() const throw()
  {   
    size_t sz=0; _CELLTYPE_* cl=_top;
    while(cl !=0) {
      ++sz; 
      cl=cl->_next;
    }
    return sz;
  }

  /// get top cell 
  _CELLTYPE_* top() const throw()
  { return _top; }

  /// insert a cell at top
  void push(_CELLTYPE_* c) throw()
  { c->_next=_top; _top=c; }       

  /// pop head 
  _CELLTYPE_* pop()  throw()
  {  
    _CELLTYPE_* c=_top;
    if(c !=0) { _top=_top->_next; c->_next=0; }
    return c;
  }  
};


//******************************************************************
//******************************************************************


// forward declaration of single lonked queue to set  friends 
template< class _CELLTYPE_  > class List;


 
/*!  \ingroup Glib
   \brief Single linked list    (thread unsafe)

  The list descriptor is linked to the first and last elements.
  We can define a list of Linkable cells  :
  \code 
  List<X> Q_of_X;
  \endcode  
  \warning In deletion, a list must be not empty. In insertion, cell
  must be defined ( not 0)
*/
template < class _CELLTYPE_ > 
class List {
protected:
  /// head
  _CELLTYPE_* _front;
  /// tail 
  _CELLTYPE_* _back;

public: 
  /// Constructor of an empty list
  List()
   : _front(0),_back(0)
  {}

  /// Destructor. The list must be empty
  ~List()throw (PosixError)
  {
    ASSERT( empty(), 
      PosixError(SUFX( "[List :] destroy a non empty list ")));
  }

  /// check empty list
  bool empty()const throw()
  { return _front ==0; }
   
  ///  get list size
  int size() const throw()
  {   
    int sz=0; _CELLTYPE_* cl=_front;
    while(cl){sz++;cl=cl->_next;}
    return sz;
  }

  /// get first cell 
  _CELLTYPE_* front()const throw()
  { return _front; }

  /// get last cell
  _CELLTYPE_* back()const throw()
  { return _back; }

  /// insert a cell at tail 
  void push_back(_CELLTYPE_* c) throw()
  {
    if (!_front){ 
      _front=c;_back=c;  c->_next=0;
    }else{
      _back->_next=c; _back=c;c->_next=0;
    }
  }    

  /// insert a cell  at head 
  void push_front(_CELLTYPE_* c) throw()
  {
    if (!_front){ 
      _front=_back=c;  c->_next;
    }else{
      c->_next=_front;  _front=c;
    }
  }   

  /// erase head 
  void pop_front()  throw()
  { pull_front(); return; }  

  /// extract head 
  _CELLTYPE_*  pull_front() throw() 
  {
    _CELLTYPE_* c=_front;
    if(!(_front=_front->_next))_back=0;// no more item 
    c->_next=0;
    return c;
  } 
 
   
  /**\brief  append a list to another  
      make the appended list empty
  */
  void append(List<_CELLTYPE_>* q) throw()
  {
    if(q->_front) {
      if(_front) {
        _back->_next=q->_front;
        _back=q->_back;
      } else {
        _front=q->_front;
        _back=q->_back;
      }
      q->_front=q->_back=0;
    }
  }
  
  /// swap 2 lists   
  void swap(List < _CELLTYPE_> * q) throw()
  { 
    _CELLTYPE_* x;
    x=q->_front;q->_front=_front;_front=x;
    x=q->_back;q->_back=_back;_back=x;
  }
};


//******************************************************************
//******************************************************************


// forward declaration of queue to set  friends 
template< class _CELLTYPE_  > class Queue;
template < class _CELLTYPE_ > class Queable;


// \ingroup Glib
/*! \brief Queueable class
  To be en(de)queueable in a a Queue, a cell  must be instance of a class derived
  from class Queable. A queueable cell may be in at most one queue.
  A queueable class must be declared as follow :
  \code 
    class X : public Queable<X> {
      ..; members etc...
    };
  \endcode 
*/
template < class _CELLTYPE_ >  
class Queable : public Linkable<_CELLTYPE_> {
protected:
  _CELLTYPE_* _pred;
  _CELLTYPE_* _next;
public :
  /// Constructor
  Queable()
   : Linkable<_CELLTYPE_>(),_pred(0)
  {}

  /// Destructor. The cell must not in a queue
  ~Queable()throw (PosixError) {
    ASSERT( !_next && !_pred, 
        PosixError(SUFX( "[Queuable :] destroy a enqueued  cell")));
  }

  /// get predecessor 
  _CELLTYPE_* pred()const throw()
  { return  _pred; }

  friend class Queue<_CELLTYPE_>; 
};
 

//  \ingroup Glib
/*!    \brief Double linked list    (thread unsafe)
  The queue descriptor is linked to the first and last elements.
  We can define a queue of Queable cells  :
  \code 
    Queue<X> Q_of_X;
  \endcode  
  \warning In deletion, queue must be not empty. In insertion, cell
  must be defined ( not 0)
*/
template < class _CELLTYPE_ > 
class Queue {
protected:
  ///   head
  _CELLTYPE_* _front;
  ///  tail 
  _CELLTYPE_* _back ;              

public : 
  /// Constructor of an empty queue
  Queue()
   :_front(0),_back(0)
  {}

  /// Destructor. The queue must be an empty queue
  ~Queue() throw (PosixError) {
//TG    return ;
    ASSERT( empty(), 
      PosixError(SUFX( "[Queue :] destroy a non empty queue ")));
  }

  /// check empty queue
  bool empty()const throw()
  { return !_front; } 

  /// clear queue
  void clear()
  { if (!empty()) while (pull_front() !=0) ; }

  /// get queue size
  int  size() const throw()
  {   
    int sz=0; _CELLTYPE_* cl=_front;
    while(cl){ sz++;cl=cl->_next; } return sz;
  }
  
  /// get first cell 
  _CELLTYPE_* front()const throw()
  { return _front; }

  /// get last cell
  _CELLTYPE_* back()const throw()
  { return  _back; }

  /// insert a cell at tail 
  void push_back(_CELLTYPE_* c) throw()
  {
    if (!_front){ 
      _front=c;_back=c;  c->_next=c->_pred=0;
    }else{
      c->_pred=_back; c->_next=0; _back->_next=c; _back=c;
    }
  }    

  /// insert a cell  at head 
   void push_front(_CELLTYPE_* c) throw()
   {
    if (!_front){ 
      _front=_back=c;  c->_next=c->_pred=0;
    }else{
      c->_next=_front; c->_pred=0; _front->_pred=c; _front=c;
    }
  }   

  /// erase head 
  void pop_front() throw()
  { pull_front(); return; }

  /// extract head 
  _CELLTYPE_* pull_front() throw() 
  {
    _CELLTYPE_* c=_front;
    if(!(_front=_front->_next))_back=0;// no more item 
    else _front->_pred=0;
    c->_next=c->_pred=0;
    return c;
  } 

  /// erase tail 
  void pop_back() throw() 
  { pull_back(); return; }

  /// extract tail 
  _CELLTYPE_* pull_back() throw() 
  { _CELLTYPE_* c=_back;
    if(!(_back=_back->_pred)) _front=0;// no more item 
    else _back->_next=0;
    c->_next=c->_pred=0;
    return c;
  } 
   
  /** \brief insert \b bf before \b af : 
    \warning \b af  must be in this queue
  */
  void insert(_CELLTYPE_* af,_CELLTYPE_* bf) throw()
  {
    if (_front==af)_front=bf; 
    else af->_pred->_next=bf;
    bf->_pred=af->_pred; bf->_next=af; af->_pred=bf;
  } 

  /** \brief substitute \b new to \b old in queue: 
      \warning \b old must be in this queue
  */
  void replace(_CELLTYPE_*old,_CELLTYPE_* nw) throw() 
  {
    if(_front==old)_front=nw;
    else old->_pred->_next=nw;
    if(_back==old) _back=nw;
    else old->_next->_pred=nw;
    nw->_pred=old->_pred; nw->_next=nw->_next;
    old->_next=old->_pred=0;
  }

  /** \brief retract a cell in queue 
    \warning \b c  must be in this queue
  */
  void erase(_CELLTYPE_* c) throw() 
  {
    if (_front==c) _front=c->_next;  
    if (_back==c) _back=c->_pred;  
    if(c->_next)c->_next->_pred=c->_pred;
    if(c->_pred)c->_pred->_next=c->_next;
    c->_next=c->_pred=0;
  }   

  /**\brief  append a  queue to another
    make the appended queue empty
  */
  void append(Queue<_CELLTYPE_>* q) throw()
  {
    if(q->_front) {
      if(_front) {
        _back->_next=q->_front;
        q->_front->_pred=_back;
        _back=q->_back;
      }else{
        _front=q->_front;
        _back=q->_back;
      }
      q->_front=q->_back=0;
    }
  }

  /// swap 2 queues   
  void swap(Queue<_CELLTYPE_>* q) throw() 
  { 
    _CELLTYPE_* x;
    x=q->_front; q->_front=_front; _front=x;
    x=q->_back; q->_back=_back; _back=x;
  }
};


//******************************************************************
//******************************************************************



// \ingroup Glib
/*!    \brief Allocator of fixed size objects  with cache for fast allocation (thread unsafe)
  A list  registers freed blocks until a given theshold. 
  Allocation tries to found a block in the list before using standard allocation.
  Deallocation puts a block in the list until list size is equal to threshold. At that time
  normal deallocation is done.
  We can define a  new allocator  of  cells  :
  \code 
    CellAllocator<X> A_of_X;
  \endcode  
*/

template < class T > class CellAllocator  {
protected:
   class Blk : public Linkable<Blk>{
   public :
     Blk(){};
     ~Blk(){};
   };
   typedef Blk* BlkRef;
   StackList<Blk > _freel;
       int _limit;
       int _ct;
public :
   /** Constructor with a given threshold
*/
   CellAllocator (int threshold):_freel(),_limit(threshold), _ct(0){ };  
  /** default Constructor with very big threshold
*/
   CellAllocator ():_freel(),_limit(0), _ct(0){ }; 
   /** \brief  destructor

    free all cached blocks 
   */
   ~CellAllocator (){
        while(! _freel.empty()) {  ::free( (void*)(  _freel.pop()));} 
       
   }
  /** init with a given threshold
*/
  void init(int threshold ){ _limit=threshold;}

   /** \brief Alloc a block of type T. Tries to get it from cached list else normal allocation
   */
   T* alloc(){ void * b;
    //std::cout << "ALLOC CELL  "<< _ct <<" "<< sizeof(T) << std::endl;
     if(_ct>0){ _ct--; b=_freel.pop(); }
     else{ b= ::malloc(sizeof(T)); }
    //std::cout << "ALLOC CELL "<< _ct <<" ("<<sizeof(T) << ","<< b <<")"<<  std::endl;
    return new(b) T();
   };
   /**  \brief Free a block of type T. Put in cached blocks if cache size is less than threshold else deallocate 
 */
   void free( T* b){   
  // std::cout << "FREE CELL "<<  b << std::endl;  
     if(_ct<_limit){ _ct++; _freel.push( BlkRef(b));}
     else ::free((void*)(b));
   };
};



//******************************************************************
//******************************************************************



/*! 
   \brief Allocator of fixed size block  with cache for fast allocation   (thread unsafe)

  An allocator manages blocks of constantt size. This size is fixed when an allocator object is build. A list registers freed blocks until a given theshold.  Allocation tries to found a block in the list before using standard allocation. Deallocation puts a block in the list until list size is equal to threshold. At that time, normal deallocation is done.

  Warning : no control of size is done

*/

template < class T >
class BlockAllocator  {
protected :
   class Blk : public Linkable<Blk>{
   public :
     Blk(){};
     ~Blk(){};
   };
   typedef Blk* BlkRef;
   StackList<Blk > _freel;
   int _bsz;
   int _limit;
   int _ct;
public :
/** default Constructor 
*/
   BlockAllocator ():_freel(),_bsz(0),_limit(0), _ct(0){};
   /// Constructor with a given threshold
   BlockAllocator (size_t block_size,int threshold):
      _freel(),_bsz(block_size),_limit(threshold), _ct(0){};  
  ///  Constructor with very big threshold
   BlockAllocator (size_t block_size):
    _freel(),_bsz(block_size),_limit(0), _ct(0){ }; 
   /** \brief  destructor. Free all cached blocks 
   */
   ~BlockAllocator (){
        while(! _freel.empty()) {  ::free( (void*)(  _freel.pop()));} 
   };
  /// init with a given threshold and size
  void init(size_t block_size,int v){ _bsz=block_size;_limit=v;_ct=0;}

   /** \brief Alloc a block. Tries to get it from cached list else normal allocation
   */
  T * alloc(){void * b;
    //std::cout << "ALLOC BLOCK  "<< _ct <<" "<< _bsz << std::endl;
     if(_ct>0){ _ct--;b= (void*)_freel.pop();}
     else {
      b= ::malloc(_bsz);
    }
    //std::cout << "ALLOC BLOCK "<< _ct <<" ("<< _bsz << ","<< b <<")"<<  std::endl;
    return new(b) T();
  };
   /**  \brief Free a block. Put in cached blocks if cache size is less than threshold else deallocate 
 */
   void free( T *b){ 
    //std::cout << "FREE BLOCK "<<  b << std::endl;  
     if(_ct<_limit){ _ct++; _freel.push( BlkRef(b));}
     else ::free((void*)(b));
   };
};


/*! 
   \brief Allocator of fixed size byte block  with cache for fast allocation   (thread unsafe)

  An allocator manages blocks of constantt size. This size is fixed when an allocator object is build. A list registers freed blocks until a given threshold.  Allocation tries to found a block in the list before using standard allocation. Deallocation puts a block in the list until list size is equal to threshold. At that time, normal deallocation is done.

  Warning : no control of size is done

*/


class ByteAllocator  {
protected :
  class Blk : public Linkable<Blk>{
  public :
    Blk(){};
    ~Blk(){};
  };
  typedef Blk* BlkRef;
  StackList<Blk > _freel;
  int _bsz;
  int _limit;
  int _ct;
public :
  /// Default constructor 
  ByteAllocator ():_freel(),_bsz(0),_limit(0),_ct(0){}
  /// Constructor with a given threshold
  ByteAllocator (size_t block_size,int threshold)
   : _freel(),_bsz(block_size),_limit(threshold), _ct(0){}
  /// default Constructor with very big threshold
  ByteAllocator (size_t block_size)
   : _freel(),_bsz(block_size),_limit(0), _ct(0){ }
  /** \brief  destructor. Free all cached blocks 
  */
  ~ByteAllocator()
  { while(! _freel.empty()) {  ::free( (void*)(  _freel.pop()));} }
  /// init with a given threshold and size
  void init(size_t block_size,int threshold)
  { _bsz=block_size;_limit=threshold;_ct=0; }

  /** \brief Alloc a block of type IOVectBuffer. Tries to get it from cached list else normal allocation
  */
  void * alloc() { 
    void * b;
  //  std::cout<< "ALLOC BYTE "<< _ct <<" "<< _bsz << std::endl;
     if(_ct>0) { _ct--;b=(void*) _freel.pop();}
     else {
      b=  ::malloc(_bsz);
    }
  //  std::cout << "ALLOC BYTE "<< _ct <<" ("<< _bsz << ","<< b <<")"<<  std::endl;
    return b;
  }
  /**  \brief Free a block. Put in cached blocks if cache size is less than threshold else deallocate 
  */
  void free( void *b) { 
   //  std::cout << "FREE BYTE "<<  b << std::endl;   
    if(_ct<_limit){ _ct++; _freel.push( BlkRef(b));}
    else ::free(b);
  }
};





}; // end namespace
          
#endif /* _PTH_BASIC_H */
