// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_pthreadsync.h,v $ 
//Copyright(c)'2001 by Apache Team // see the copyright file 
// $Id: pth_pthreadsync.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $ 
//==========================================================================

/*!
\file  
\brief  Posix thread for Inuktitut0++ INRIA 
\author  J. Briat T. Gautier 
\date 2001

This file contains the wrappers to :
  - posix locks PTH::Mutex
  - posix condition PTH::Condition
  - part of the posix thread interface PTH::Thread PTH::ThreadModel
  - a lexical scoped lock mechanism CRITICAL()

It provides operators on these objects throwing Posix Exceptions.
\sa pth_error.h PTH::PosixError
*/
#ifndef _PTH_PTHREADSYNC_H
#define _PTH_PTHREADSYNC_H

// the namespace identifier
#include <bits/libc-lock.h>
#include <inuk/pth_ns.h>
#include <inuk/pth_dep.h>
#include <inuk/pth_macro.h>
#include <inuk/pth_error.h>
#include <sys/time.h>
#include <inuk/pth_pthreadsync_notrace.h>


#if defined(PTH_USE_PTHREAD_DARWIN) 
#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_NORMAL
#define MUTEX_TYPE_RECURSIVE PTHREAD_MUTEX_RECURSIVE

#elif defined(PTH_USE_PTHREAD_LINUX)
#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_FAST_NP
#define MUTEX_TYPE_RECURSIVE PTHREAD_MUTEX_RECURSIVE_NP

#elif defined(PTH_USE_PTHREAD_LINUX_IA64)
#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_FAST_NP
#define MUTEX_TYPE_RECURSIVE PTHREAD_MUTEX_RECURSIVE_NP

#elif defined(PTH_USE_PTHREAD_IRIX)
#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_NORMAL
//#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_SPINBLOCK_NP
#define MUTEX_TYPE_RECURSIVE PTHREAD_MUTEX_RECURSIVE
//#define COND_TYPE PTHREAD_PROCESS_SHARED
#define COND_TYPE PTHREAD_PROCESS_PRIVATE

#else
#error UNABLE TO DEFINE POSIX THREAD ATTRIBUTES
#endif


namespace PTH { 

class Pth; // forward decl

//! @defgroup Synchro Synchronization facilities
//! @brief All the classes to use for synchro.

// -------------------------------------------------------------------------
// class Mutex
//! @brief Exclusive code execution.
//! @ingroup Synchro
// -------------------------------------------------------------------------

//!
//! This class implements interface for Posix mutex.
//! @warning Be careful to unlock a mutex in case of exception throwing. 
//! @warning Use class PTH::CriticalGuard to do this safely.

class Mutex : protected NTMutex {
public:
  //! Constructor. The mutex is NOT locked when returning.
  Mutex() throw () 
  {}

  //! Virtual destructor
  ~Mutex() throw (PosixError) 
  { } 
  
  //! Initialize a  recursive or flat (default case) mutex in unlock state.
  //! @param recursive if true the mutex is a recusrive one.
  void init(bool recursive=false) throw (PosixError);
  
  //! Explicit destroy. Must be called before the destructor is called. 
  //! @throw PosixError if the mutex is not initialized
  //! @throw PosixError if the mutex is locked
  void destroy () throw (PosixError);

  //! Lock request. Wait until access is granted.
  //! @throw PosixError if the mutex is not initialized
  void lock() throw( PosixError);

  //! Lock if possible.
  //! @throw PosixError if the mutex is not initialized
  //! @returns false if the mutex already owned
  bool trylock() throw (PosixError);

  //! Unlock request.
  //! @throw PosixError if the mutex is not initialized
  //! @throw PosixError if the current thread is not the owner
  void unlock() throw( PosixError);


  //! @name Methods returning an error code instead of throwing exceptions
  //@{

  //! Initialize a  recursive or flat (default case) mutex in unlock state.
  //! @param recursive if true the mutex is a recusrive one.
  //! @return 0 if OK
  int initnothrow(bool recursive = true ) throw ();

  //! Explicit destroy. Must be called before the destructor is called. 
  //! @return 0 if OK, <>0 if mutex is locked or not initialized.
  int destroynothrow () throw ();

  //! Lock request. Wait until access is granted.
  //! @return 0 if OK, <>0 if the mutex is not initialized
  int locknothrow() throw ();

  //! Unlock request.
  //! @return 0 if OK, <>0 if the mutex is not initialized or not owned by 
  //! the current thread.  
  int unlocknothrow() throw ();

  //!  Lock if possible.
  //! @throw if the mutex is not initialized
  //! @returns 0 if OK, EBUSY if the mutex already owned, else the mutex is not
  //! initialized.
  int trylocknothrow() throw ();
  //@}

  //! @name Methods intended to the tracer
  //@{
  void init_notrace() throw ();   

  void lock_notrace() throw( PosixError);

  void unlock_notrace() throw( PosixError);
  //@}

protected: 
  /// static values initialisation 
  static void init_static() throw (PosixError); 
  
  friend class Condition;
  friend class Pth;
};


// -------------------------------------------------------------------------
// class SpinMutex
//! @brief Spin locks
//! @ingroup Synchro
// -------------------------------------------------------------------------

//! This class implements interface for Posix mutex using spin lock
//! technics.
//! @warning Be careful to unlock a mutex in case of exception throwing. 
//! @warning Use class PTH::CriticalGuard to do this safely.

class SpinMutex : public Mutex {
public:
  //! Constructor. The mutex is NOT locked when returning.
  SpinMutex() throw ()
   : Mutex(), _state(0), _spin_count(100)
#ifdef PTH_DEBUG
   , _isinit(false)  
#endif
  {}

  //! Virtual destructor
  ~SpinMutex() throw (PosixError) {
#ifdef PTH_DEBUG
    _isinit =false;  
#endif
  } 
  

  //! @name Methods returning an error code instead of throwing exceptions
  //@{

  //! Initialize a spin mutex with a spin count value.
  //! @param spin_count number of spin to do before releasing processor.
  //! @param recursive if true the mutex is a recusrive one.
  //! @return 0 if OK
  int initnothrow( int spin_count = 100, bool recursive=false ) throw ();

  //! Lock request. Wait until access is granted.
  //! @return 0 if OK, <>0 if the mutex is not initialized
  int locknothrow() throw ();

  //! Unlock request.
  //! @return 0 if OK, <>0 if the mutex is not initialized or not owned by 
  //! the current thread.  
  int unlocknothrow() throw ();
  
  //!  Lock if possible.
  //! @throw if the mutex is not initialized
  //! @returns 0 if OK, EBUSY if the mutex already owned, else the mutex is not
  //! initialized.
  int trylocknothrow() throw ();
  //@}

  //! Initialize a spin mutex with a spin count value.
  //! @param spin_count number of spin to do before releasing processor.
  //! @param recursive if true the mutex is a recusrive one.
  void init( int spin_count = 100, bool recursive=false) throw (PosixError);
  
  //! Lock request. Wait until access is granted.
  //! @throw PosixError if the mutex is not initialized
  void lock() throw( PosixError);

  //! Lock if possible.
  //! @throw PosixError if the mutex is not initialized
  //! @returns false if the mutex already owned
  bool trylock() throw (PosixError);

  //! Unlock request.
  //! @throw PosixError if the mutex is not initialized
  //! @throw PosixError if the current thread is not the owner
  void unlock() throw( PosixError);

protected: 
  int             _state; //!< state of mutex: 1 locked, 0 unlocked
  int             _spin_count; //!< spin count defined at init
#ifdef PTH_DEBUG
  bool _isinit;
#endif
 private:
  friend class Condition;
  friend class Pth;
};
 

// -------------------------------------------------------------------------
// class CriticalGuard
//! @brief Transparent mutual exclusion facility.
//! @ingroup Synchro
// -------------------------------------------------------------------------

//! This class implements a  transparent local scoping usage 
//! of mutex lock. An instance of this class is used to lock a block 
//! expression (i.e. between { and }).
//! The instance delaration must appear at the beginning of the block.
//! At the initialization of such an object, the associated mutex is locked 
//! and is unlocked at the destruction of the object (at the end of its 
//! scope as a C++ variable). Note that the unlock method is alway called 
//! even if an exception is thrown.
//! Find examples of use @ref pth_critical_ex "here".
//! @see CRITICAL
//! @see PTH::TCriticalGuard

class CriticalGuard {
private :
  PTH::Mutex * _mtx_;
public : 
  CriticalGuard (Mutex &mtx);
  CriticalGuard (Mutex *mtx);
  ~CriticalGuard ();
};

// -------------------------------------------------------------------------
// class TCriticalGuard
//! @brief Generic CriticalGuard facility.
//! @ingroup Synchro
// -------------------------------------------------------------------------

//! This class extend the usage of CriticalGuard object to 
//! any kind of object having a "lock/unlock" interface.
//! Find examples of use @ref pth_tcritical_ex "here".
//! @see PTH::CriticalGuard

template<class LOCKABLE>
class TCriticalGuard {
private :
  LOCKABLE * _mtx_;
public : 
  TCriticalGuard (LOCKABLE &mtx);
  TCriticalGuard (LOCKABLE *mtx);
  ~TCriticalGuard ();
};


//! @def  CRITICAL()
//! This macro can be use in place of PTH::CriticalGuard 
//! @see PTH::CriticalGuard
#define CRITICAL(MTX) PTH::CriticalGuard  _CONCAT(_cg, __LINE__) (MTX)



// -------------------------------------------------------------------------
// class Condition
//! @brief Posix conditions.
//! @ingroup Synchro
// -------------------------------------------------------------------------

//! This class implements interface to posix condition variable.
//! Find examples of use @ref pth_cond_ex "here".

class Condition {
public:   
  //! Constructor
  Condition() throw ()
  {}
  //! Destructor. 
  //! @throws PosixError if there are waiting threads.
  virtual ~Condition() throw (PosixError) 
  { } 

  //! Initialize a  condition. Must be called before use.
  //! @throws PosixError if there is a problem.
  void init() throw (PosixError); 

  //! Explicit destroy. Must be called before destructor.
  //! @throws PosixError if there is waiting threads.
  void destroy () throw (PosixError);

  //! Wait for condition. The thread is blocked until the condition is 
  //! signaled by a call to signal or broadcast. The mutex is automatically 
  //! released and re-acquired during the call. 
  //! @param mutex The mutex protecting the condition variable.
  //! @throws PosixError if the condition is not initialized.
  //! @throws PosixError if the mutex is not initialized.
  //! @throws PosixError if the mutex is not owned.
  void wait( Mutex& mutex ) throw (PosixError);

  //! Relative bounded wait for condition. 
  //! The thread is blocked until the condition is 
  //! signaled by a call to signal or broadcast or if the delay expires.
  //! The mutex is automatically released and re-acquired during the call. 
  //! @param mutex The mutex protecting the condition variable. 
  //! @param milli length in milliseconds of timeout
  //! @param nano  additional length in nanoseconds of timeout
  //! @throws PosixError if the condition is not initialized.
  //! @throws PosixError if the mutex is not initialized.
  //! @throws PosixError if the mutex is not owned.
  //! @return true if awaken by timeout.
  bool wait( Mutex& mutex , long milli, int nano=0) throw(PosixError); 

  //! Absolute bounded wait for condition. 
  //! The thread is blocked until the condition is 
  //! signaled by a call to signal or broadcast or if the absolute specified 
  //! time is reached. 
  //! The mutex is automatically released and re-acquired during the call. 
  //! @param mutex The mutex protecting the condition variable. 
  //! @param abstime absolute time that cannot be exceeded.
  //! @throws PosixError if the condition is not initialized.
  //! @throws PosixError if the mutex is not initialized.
  //! @throws PosixError if the mutex is not owned.
  //! @return true if abstime is reached
  bool wait( Mutex& mutex, const struct timespec& abstime ) throw (PosixError);

  //! Signal condition to one blocked thread. 
  //! If no thread is blocked for the condition this call has no effect.
  //! @throws PosixError if the condition is not initialized.
  void signal() throw (PosixError);

  //! Broadcast condition to all threads. 
  //! If no thread is blocked for the condition this call has no effect.
  //! @throws PosixError if the condition is not initialized.
  void broadcast() throw (PosixError); 
    

  //! @name Methods returning an error code instead of throwing exceptions
  //@{

  
  //! Initialize a  condition. Must be called before use.
  //! @return <>0 if there is a problem.
  int initnothrow() throw ();

  //! Explicit destroy. Must be called before destructor.
  //! @return <>0 if there are waiting threads.
  int destroynothrow () throw ();

  //! Wait for condition. The thread is blocked until the condition is 
  //! signaled by a call to signal or broadcast. The mutex is automatically 
  //! released and re-acquired during the call. 
  //! @param mutex The mutex protecting the condition variable.
  //! @return <>0 if the condition is not initialized.
  //! @return <>0 if the mutex is not initialized.
  //! @return <>0 if the mutex is not owned.
  int waitnothrow( Mutex& mutex ) throw ();

  //! Relative bounded wait for condition. 
  //! The thread is blocked until the condition is 
  //! signaled by a call to signal or broadcast or if the delay expires.
  //! The mutex is automatically released and re-acquired during the call. 
  //! @param mutex The mutex protecting the condition variable. 
  //! @param milli length in milliseconds of timeout
  //! @param nano  additional length in nanoseconds of timeout
  //! @return ETIMEOUT if awaken by timeout.
  //! @return <>0 and <>ETIMEOUT if the condition is not initialized.
  //! @return <>0 and <>ETIMEOUT if the mutex is not initialized.
  //! @return <>0  and <>ETIMEOUT if the mutex is not owned.
  int waitnothrow( Mutex& mutex ,long milli, int nano=0) throw ();

  //! Absolute bounded wait for condition. 
  //! The thread is blocked until the condition is 
  //! signaled by a call to signal or broadcast or if the absolute specified 
  //! time is reached. 
  //! The mutex is automatically released and re-acquired during the call. 
  //! @param mutex The mutex protecting the condition variable. 
  //! @param abstime absolute time that cannot be exceeded.
  //! @return ETIMEOUT abstime is reached.
  //! @return <>0 and <>ETIMEOUT if the condition is not initialized.
  //! @return <>0 and <>ETIMEOUT if the mutex is not initialized.
  //! @return <>0  and <>ETIMEOUT if the mutex is not owned.
  int waitnothrow( Mutex& mutex , const struct timespec& abstime ) throw ();

  //! Signal condition to one blocked thread. 
  //! If no thread is blocked for the condition this call has no effect.
  //! @return <>0 if the condition is not initialized.
  int signalnothrow() throw ();

  //! Broadcast condition to all threads. 
  //! If no thread is blocked for the condition this call has no effect.
  //! @return <>0 if the condition is not initialized.
  int broadcastnothrow() throw ();  

  //@}

protected:
  pthread_cond_t _condition;   //!< the posix condition

private:
  friend class Pth;
  static pthread_condattr_t default_attr; 
  static void init_static() throw (PosixError); 
  // friend void initialize(int* , char***) throw (PosixError);
};


// -------------------------------------------------------------------------
// class Key
//! Thread private attribute.
// -------------------------------------------------------------------------


//! This class provides an interface to posix pthread specific
//! attributes.
//! It only includes operators for :
//!  - getting specific data key create() with a value destructor function 
//!    called at the end of the thread
//!  - setting/getting specific data set() get()
template < class T> class Key : public NTKey<T> { 
};


#define PTH_CLEANER(HH,AA) PTH::Cleaner _CL_(HH,AA)

// -------------------------------------------------------------------------
// class Cleaner
//! Cleaner for posix threads
// -------------------------------------------------------------------------

 
//! This class Cleaner provides an  interface to posix cleanup handler.
//! The role of a cleaner is to free the ressources (mutex, memory 
//! regions) of a thread when it terminates not normally. 
//! CleanUp handlers MUST be allocated on stack and MUST have block 
//! or procedure scope.

class Cleaner {
protected :
#if defined(PTH_USE_PTHREAD_LINUX) || defined(PTH_USE_PTHREAD_LINUX_IA64)
  struct _pthread_cleanup_buffer _clup;
#elif defined(PTH_USE_PTHREAD_DARWIN)
  struct _pthread_handler_rec __handler;
#endif
  // TODO :
public :
  //! Register an cleanup handler.
  //! @param hdl Handler to be called if the thread terminates before the
  //! destructor is called.
  Cleaner(void (*hdl)(void*), void*  v);

  //! Unregister a cleanup handler. The handler will no longer be called on
  //! thread termination.
  ~Cleaner();
};

// -------------------------------------------------------------------------
// class Thread
//! POSIX threads
// -------------------------------------------------------------------------

//! This class provides set of static functions as interface to posix pthread 
//! It only includes functions for:
//! - cancellation disabling (Thread::disableCancel())
//! - yielding control (Thread::yield())
//! - sleep during a period of time (Thread::sleep())
//! - creating Pthread in joinable or detached mode (Thread::create())
//!	- joining pthread started in joinable mode (Thread::join())
//! - getting the current thread id (Thread::getSelf())

typedef NTThreadId ThreadId;

#define PTH_DISABLE_CANCEL  PTH::Thread::disableCancel()

class Thread {

public:
  typedef pthread_attr_t * Kind; //!< attribute type of a thread
  static  Kind DETACHED;         //!< To create a thread as detached
  static  Kind JOINABLE;         //!< To create a thread as joinable

  //! Cstor
  Thread() throw() {}
  //! Dstor
  ~Thread() throw() {}

  //! Current thread yield the processor if its priority is lower 
  //! than some ready threads
  static void yield() throw ();

  //! Current thread yield the processor for an given delay
  static int sleep( long milli, long nano) throw ();

  //! Current thread disable automatic cancellation
  static  int disableCancel( ) throw();

  //! Terminate the current thread.
  //! @param x pointer on a return value (posix type) if exists
  static void exit( void*x=0)throw (){ pthread_exit(x);}

  //! terminate the current thread with a pointer on a value 
  //! (cast to posix type) if exists
  //! @param x pointer on a return value (posix type) if exists
  template <class Tr >
  static void exit( Tr*x=0)throw (){ pthread_exit((void*)x);}

  //! Thread creation (posix type).
  //! Create a thread (posix type). 
  //! @param thid pointer to the Id givent to the created thread.
  //! @param f entry point of thr thread.
  //! @param a argument to give to f.
  //! @param k kind of thread (default is DETACHED)
  //! @throw PosixError if it fails
  static void create(ThreadId * thid, void  *(*f)(void *),void * a, 
                     Kind k= DETACHED) throw(PosixError);

  //! Thread creation (cast to posix type).
  //! Create a detached thread (cast argument to posix type).
  //! @param thid pointer to the Id givent to the created thread.
  //! @param f entry point of thr thread.
  //! @param a argument to give to f.
  //! @param k kind of thread (default is DETACHED)
  //! @throw PosixError if it fails
  template <class Tr , class Ta> 
  static void create(ThreadId * thid, Tr *(*f)(Ta *),Ta* a, Kind k= DETACHED ) 
    throw(PosixError);
//  { create(th, (void  *(*)(void *))f , (void *)a, k); }


  //! Thread junction (posix type). 
  //! Join a joinable thread (posix type). 
  //! @param thid Id of thread to join
  //! @param v ?? 
  //! @throws PosixError if it fails
  static void join(ThreadId  thid,void **v=0) throw (PosixError);

  //! Thread junction (cast to posix type). 
  //! Join a joinable thread (cast argument to posix type). 
  //! @param thid Id of thread to join
  //! @param v ?? 
  //! @throws PosixError if it fails
  template <class Tr >
  static void join(ThreadId  thid, Tr** v=0) throw (PosixError)
  { join( thid,(void**)v); }


  //! Get current thread id.
  //! @return current thread id.
  static ThreadId  getSelf()throw ()
  { return ::pthread_self(); }


  //! Thread creation (posix type).
  //! Create a detached thread (posix type). 
  //! @param thid pointer to the Id givent to the created thread.
  //! @param f entry point of thr thread.
  //! @param a argument to give to f.
  //! @param k kind of thread (default is DETACHED)
  //! @return <>0 if it fails
  static int createnothrow(ThreadId * thid, void  *(*f)(void *),void * a,
                           Kind k= DETACHED) throw();

  //! Thread creation (cast to posix type).
  //! Create a detached thread (cast argument to posix type). 
  //! @param thid pointer to the Id givent to the created thread.
  //! @param f entry point of thr thread.
  //! @param a argument to give to f.
  //! @param k kind of thread (default is DETACHED)
  //! @return <>0 if it fails
  template <class Tr , class Ta> 
  static int createnothrow(ThreadId * th, Tr *(*f)(Ta *), Ta* a, Kind k= DETACHED) throw();

  //! Thread junction (posix type). 
  //! Join a joinable thread (posix type). 
  //! @param thid Id of thread to join
  //! @param v ?? 
  //! @return <>0 if it fails
  static int joinnothrow(ThreadId  thid,void **v=0) throw();

  //! Thread junction (cast to posix type). 
  //! Join a joinable thread (cast argument to posix type). 
  //! @param thid Id of thread to join
  //! @param v ?? 
  //! @return <>0 if it fails
  template <class Tr >
  static int joinnothrow(ThreadId  thid, Tr** v=0) throw()
  { return joinnothrow( thid,(void**)v); }



private:
  static  pthread_attr_t detached_attr;
  static  pthread_attr_t joinable_attr;
  /// static values initialisation 
  static void init_static() throw (PosixError); 
  friend class Pth;
};


/** 
\brief These class  provide a wrapper class fromC++ object to  posix pthread 
  needed for  Inuktitut0 threads .

Any classes providing a void run() method may be executed as a thread.

It inherits of the Thread class operators. It  includes  operators :
	- to start a thread and encapsulate its id. PTH::ThreadModel::start() PTH::ThreadModel::startnothrow()
	- to join a thread PTH::ThreadModel::join() PTH::ThreadModel::joinnothrow()
	- to get the Posix thread id PTH::ThreadModel::getThreadId()
*/
class BasicThreadModel : public Thread {
static void* runIt( void *s);
public:
  /* Need */
  virtual ~BasicThreadModel() {}

  /* Needed to make the right call to object XX
     to validate translation : 
     - start: ThreadModel* -> XX*
     - runIt: XX* a, a->run
     Cost a virtual method call
  */
  virtual void* run() =0;
  
  /// Native thread id . Is it  really used ??
  ThreadId       _threadid;    

  /// start a thread object. An exception is thrown if it fails
  void start(Thread::Kind k= Thread::DETACHED) throw(PosixError);

  /// start a thread object. An  error code is returned if it fails
  int startnothrow(Thread::Kind k= Thread::DETACHED) throw(PosixError);

  /// join the thread. An exception is thrown if it fails. Get a pointer on the result value 
  template < class Tr >
  void join( Tr** v=0) throw (PosixError)
  { Thread::join(_threadid,&v); }

  /// join the thread. An  error code is returned if it fails. Get a pointer on the result value
  template < class Tr >
  int joinnothrow( Tr** v=0) throw ()
  { Thread::join(_threadid,v); }

  /// get native  thread id
  ThreadId getThreadId()
  { return _threadid; }
};

template < class XX > 
class ThreadModel : public BasicThreadModel {
};

//class ClassicThread : public ThreadModel< ClassicThread >{};

//ClassicThread CL;



/*
 * inline definitions
 */
// -------------------------------------------------------------------------
// Mutex
// -------------------------------------------------------------------------
inline int Mutex::initnothrow(bool recursive ) throw ()
{
  int err;
#ifdef PTH_USE_TRACER
  MUTEX_INIT_0(&_mutex);
#endif
  if(recursive)
    { err = pthread_mutex_init(&_mutex, &recursive_attr); }
  else
    { err = pthread_mutex_init(&_mutex, &default_attr); }
#ifdef PTH_DEBUG
  if (err ==0) _isinit =true;
#endif
#ifdef PTH_USE_TRACER
  MUTEX_INIT_1(&_mutex);
#endif
  return err;
}

inline int Mutex::destroynothrow () throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_DESTROY_0(&_mutex);
#endif
  int err;
  err=pthread_mutex_destroy(&_mutex);
#ifdef PTH_DEBUG
  if (err ==0) _isinit =false;
#endif
#ifdef PTH_USE_TRACER
  MUTEX_DESTROY_1(&_mutex);
#endif
  return err;
}

inline int Mutex::locknothrow() throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_LOCK_0(&_mutex);
#endif
  int ret= pthread_mutex_lock(&_mutex);
#ifdef PTH_USE_TRACER
  MUTEX_LOCK_1(&_mutex);
#endif
  return ret;
}

inline int Mutex::unlocknothrow() throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_UNLOCK_0(&_mutex);
#endif
  int err;
  err= pthread_mutex_unlock(&_mutex);
#ifdef PTH_USE_TRACER
  MUTEX_UNLOCK_1(&_mutex);
#endif
  return err;
}

inline int Mutex::trylocknothrow() throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_TRYLOCK_0(&_mutex);
#endif
  int ret= pthread_mutex_trylock(&_mutex);
#ifdef PTH_USE_TRACER
  MUTEX_TRYLOCK_1(&_mutex);
#endif
  return ret;
}

inline void Mutex::init(bool recursive) throw (PosixError)
{
  int err;
  ASSERT( !(err=initnothrow(recursive)),
  PTH::PosixError(SUFX("[Mutex::init] POSIX error"), err));
}

inline void Mutex::destroy () throw (PosixError)
{
  int err;
  ASSERT( !(err=destroynothrow()),
  PTH::PosixError(SUFX("[Mutex::destroy] POSIX error"),err));
}


inline void Mutex::lock() throw( PosixError)
{
  int err;
  ASSERT( !(err= locknothrow()),
  PTH::PosixError(SUFX("[Mutex::lock] POSIX error"),err));
}

inline bool Mutex::trylock() throw (PosixError)
{
  int err;
  err= trylocknothrow();
  if (err==EBUSY) return false;
  if (err==0) return true;
  ASSERT( !err, PTH::PosixError(SUFX("[Mutex::trylock] POSIX error"),err));
  return false;
}

inline void Mutex::unlock() throw( PosixError)
{
  int err;
  ASSERT( !(err= unlocknothrow()), PosixError(SUFX("[Mutex::unlock] POSIX error"),err));
}

inline void Mutex::init_notrace() throw ()
{
  pthread_mutex_init(&_mutex,&default_attr);
}

inline void Mutex::lock_notrace() throw( PosixError)
{
  int err;
  ASSERT( !(err= pthread_mutex_lock(&_mutex)),
    PTH::PosixError(SUFX("[Mutex::lock] POSIX error"),err));

}

inline void Mutex::unlock_notrace() throw( PosixError)
{
  int err;
  ASSERT( !(err= pthread_mutex_unlock(&_mutex)),
     PTH::PosixError(SUFX("[Mutex::unlock] POSIX error"),err));
}



// -------------------------------------------------------------------------
// SpinMutex
// -------------------------------------------------------------------------
inline int SpinMutex::initnothrow( int spin_count, bool recursive) throw ()
{
  int err;
#ifdef PTH_USE_TRACER
  MUTEX_INIT_0(&_mutex);
#endif
  if(recursive)
    { err = pthread_mutex_init(&_mutex, &Mutex::recursive_attr); }
  else
    { err = pthread_mutex_init(&_mutex, &Mutex::default_attr); }
#ifdef PTH_DEBUG
  if (err ==0) _isinit =true;
#endif
  _state = 0;
  _spin_count = spin_count;
#ifdef PTH_USE_TRACER
  MUTEX_INIT_1(&_mutex);
#endif
  return err;
}

inline int SpinMutex::locknothrow() throw ()
{
  //std::cout << "IN SpinMutex::lock" << std::endl;
#ifdef PTH_USE_TRACER
  MUTEX_LOCK_0(&_mutex);
#endif
  int err;
  for (int i=0; i<_spin_count; ++i) {
    if (_state) continue;
    err = pthread_mutex_trylock(&_mutex);
    if (err == 0) {
      _state = 1;
#ifdef PTH_USE_TRACER
      MUTEX_LOCK_1(&_mutex);
#endif
      return 0;
    }
    if (err ==EBUSY)
      continue;
    return err;
  }
  err = pthread_mutex_lock(&_mutex);
  _state = 1;
#ifdef PTH_USE_TRACER
  MUTEX_LOCK_1(&_mutex);
#endif
  return err;
}

inline int SpinMutex::unlocknothrow() throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_UNLOCK_0(&_mutex);
#endif
  int err;
  _state =0;
  err= pthread_mutex_unlock(&_mutex);
#ifdef PTH_USE_TRACER
  MUTEX_UNLOCK_1(&_mutex);
#endif
  return err;
}

inline int SpinMutex::trylocknothrow() throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_TRYLOCK_0(&_mutex);
#endif
  int err= pthread_mutex_trylock(&_mutex);
  if (err ==0) {
    _state =1;
  }
#ifdef PTH_USE_TRACER
  MUTEX_TRYLOCK_1(&_mutex);
#endif
  return err;
}

inline void SpinMutex::init(int spin_count, bool recursive) throw (PosixError)
{
  int err;
  ASSERT( !(err=initnothrow(spin_count, recursive)),
  PTH::PosixError(SUFX("[Mutex::init] POSIX error"), err));
}

inline void SpinMutex::lock() throw( PosixError)
{
  int err;
  ASSERT( !(err= locknothrow()),
    PTH::PosixError(SUFX("[Mutex::lock] POSIX error"),err));
}

inline bool SpinMutex::trylock() throw (PosixError)
{
  int err;
  err= trylocknothrow();
  if (err==0) return true;
  if (err==EBUSY) return false;
  ASSERT( !err, PTH::PosixError(SUFX("[Mutex::trylock] POSIX error"),err));
  return false;
}

inline void SpinMutex::unlock() throw( PosixError)
{
  int err;
  ASSERT( !(err= unlocknothrow()), PosixError(SUFX("[Mutex::unlock] POSIX error"),err));
}



// -------------------------------------------------------------------------
// Condition
// -------------------------------------------------------------------------
inline int Condition::initnothrow() throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_INIT_0(&_condition);
#endif
  int err;
  err=pthread_cond_init(&_condition, &default_attr);
#ifdef PTH_USE_TRACER
  CONDITION_INIT_1(&_condition);
#endif
  return err;
}

inline void Condition::init() throw (PosixError)
{
  int err;
  ASSERT( !(err=initnothrow()), PosixError (SUFX("[Condition()] POSIX error"), err));
}

inline int Condition::destroynothrow()throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_DESTROY_0(&_condition);
#endif
  int err;
  err=pthread_cond_destroy( &_condition );
#ifdef PTH_USE_TRACER
  CONDITION_DESTROY_1(&_condition);
#endif
return err;
}

inline void Condition::destroy()throw (PosixError)
{
  int err;
  ASSERT( !(err=destroynothrow()),
  PosixError(SUFX( "[Condition::~()] POSIX error"),err));
}

inline int Condition::waitnothrow( Mutex& mutex )throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_0(&(mutex._mutex),&_condition,0,0);
#endif
  int ret= pthread_cond_wait( &_condition, &(mutex._mutex));
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_1(&(mutex._mutex),&_condition,0,0);
#endif
  return ret;
}

inline void Condition::wait( Mutex& mutex ) throw (PosixError)
{
  int err;
  ASSERT( !(err=waitnothrow(mutex)), PosixError(SUFX( "[Condition::wait] POSIX error"),err));
}

inline int Condition::waitnothrow( Mutex& mutex, const struct timespec& time ) throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_0(&(mutex._mutex),&_condition,time.tv_sec,time.tv_nsec);
#endif
  int r = pthread_cond_timedwait( &_condition, &(mutex._mutex),&time);
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_1(&(mutex._mutex),&_condition,time.tv_sec,time.tv_nsec);
#endif
  return r;
}

inline bool Condition::wait( Mutex& mutex, const struct timespec& time ) throw (PosixError)
{
  int r=waitnothrow(mutex,time);
  if (r ==0) return false;
  if (r == ETIMEDOUT) return true;
  ASSERT( !r,PosixError(SUFX(  "Condition::timedwait"), r));
  return false;
}

inline int Condition::waitnothrow( Mutex& mutex, long milli, int nano ) throw ()
{
  struct timespec req;
  struct timeval now;
  gettimeofday(&now,0);
  req.tv_sec  =  now.tv_sec+ milli / 1000;
  req.tv_nsec =  (now.tv_usec*1000)+ (1000000 * (milli % 1000)) + nano;
  return waitnothrow( mutex,req);
}

inline bool Condition::wait( Mutex& mutex, long milli, int nano ) throw (PosixError)
{
  int r=waitnothrow(mutex,milli,nano);
  if (r ==0) return false;
  if (r == ETIMEDOUT) return true;
  ASSERT( !r,PosixError(SUFX(  "Condition::timedwait"), r));
  return false;
}

inline int Condition::signalnothrow()throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_SIGNAL_0(&_condition);
#endif
  int ret= pthread_cond_signal(&_condition);
#ifdef PTH_USE_TRACER
  CONDITION_SIGNAL_1(&_condition);
#endif
  return ret;
}

inline void Condition::signal()throw (PosixError)
{
  int err;
  ASSERT( !(err=signalnothrow()), PosixError (SUFX("[Condition::signal] POSIX error"),err));
}

inline int Condition::broadcastnothrow()throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_BROADCAST_0(&_condition);
#endif
  int ret= pthread_cond_broadcast(&_condition);
#ifdef PTH_USE_TRACER
  CONDITION_BROADCAST_1(&_condition);
#endif
  return ret;
}

inline void Condition::broadcast()throw (PosixError)
{
  int err;
  ASSERT( !(err=broadcastnothrow()), PosixError(SUFX("[Condition::broadcast] POSIX error"),err)); 
}


// -------------------------------------------------------------------------
// Cleaner (For Posix only)
// -------------------------------------------------------------------------  
#if defined(PTH_USE_PTHREAD_LINUX) || defined(PTH_USE_PTHREAD_LINUX_IA64)
inline Cleaner::~Cleaner()
{
  _pthread_cleanup_pop(&_clup,0);
}

inline Cleaner::Cleaner( void (*hdl)(void*), void*v)
{
  _pthread_cleanup_push(&_clup,hdl,v);
}
#elif defined(PTH_USE_PTHREAD_DARWIN)
/* TG : implemented using /usr/include/pthread.h definition ! */
inline Cleaner::~Cleaner()
{
  pthread_t __self = pthread_self();
  __self->cleanup_stack = __handler.next;
  if (0) (__handler.routine)(__handler.arg);
}

inline Cleaner::Cleaner( void (*hdl)(void*), void* v)
{
  pthread_t __self = pthread_self();
  __handler.routine = hdl;
  __handler.arg = v;
  __handler.next = __self->cleanup_stack;
  __self->cleanup_stack = &__handler;
}
#endif



// -------------------------------------------------------------------------
// Thread 
// -------------------------------------------------------------------------
inline int Thread::createnothrow(ThreadId * thid, void  *(*f)(void *),void * a, Kind k) throw()
{
  int ret= pthread_create(thid, k, f, a);
#ifdef PTH_USE_TRACER
  PTH_CREATE_1(*thid);
#endif
  return ret;
}
template <class Tr , class Ta>
int Thread::createnothrow(ThreadId * th, Tr *(*f)(Ta *), Ta* a, Kind k) throw()
{ return createnothrow(th, (void  *(*)(void *))f , (void *)a, k);}

inline void Thread::create(ThreadId * thid, void  *(*f)(void *),void * a, Kind k) throw(PosixError)
{
  int err;
  ASSERT( (!(err=createnothrow(thid, f, a, k))),
  PosixError(SUFX("[Thread::create] unable to create POSIX thread "),err));
}


inline  void Thread::yield()throw ()
{
#ifdef PTH_USE_TRACER
  PTH_YIELD_0(Thread::getSelf());
#endif
  sched_yield();
#ifdef PTH_USE_TRACER
  PTH_YIELD_1(Thread::getSelf());
#endif
}


#if defined(PTH_USE_PTHREAD_DARWIN)
extern "C" int nanosleep(struct timespec* req, int );
#endif

inline int Thread::sleep( long milli, long nano) throw ()
{
#ifdef PTH_USE_TRACER
  PTH_SLEEP_0(pthread_self(),milli, nano);
#endif
#if defined(PTH_HAVE_NANOSLEEP)
  struct timespec req;
  req.tv_sec  = milli / 1000;
  req.tv_nsec = 1000000 * (milli % 1000) + nano;
  int ret=nanosleep(&req,0);
#ifdef PTH_USE_TRACER
  PTH_SLEEP_1(getSelf(),milli, nano);
#endif
  return ret;
#elif defined(PTH_HAVE_MICROSLEEP)
  unsigned int microsec = milli*1000 + nano/1000;
  usleep( microsec );
#ifdef PTH_USE_TRACER
  PTH_SLEEP_1(getSelf(),milli, nano);
#endif
  return 0;
#elif defined(PTH_HAVE_SLEEP)
  unsigned int sec = milli/1000 + nano/1000000000;
  sleep(sec);
#ifdef PTH_USE_TRACER
  PTH_SLEEP_1(getSelf(),milli, nano);
#endif
  return 0;
#else
#error UNABLE TO DEFINE SLEEP
#endif
}

inline  int Thread::disableCancel( )throw()
{
  return pthread_setcancelstate( PTHREAD_CANCEL_DISABLE ,0);
}

inline int Thread::joinnothrow(ThreadId  th, void **v) throw()
{
#ifdef PTH_USE_TRACER
  PTH_JOIN_0(th);
#endif
  int ret= pthread_join(th,v);
#ifdef PTH_USE_TRACER
  PTH_JOIN_1(th);
#endif
  return ret;
}

inline void Thread::join(ThreadId  thid, void **v) throw(PosixError)
{
  int err;
  ASSERT( (!(err=joinnothrow(thid,v))),
  PosixError(SUFX("[Thread::join] unable to join POSIX thread "),err));
}

inline void* BasicThreadModel::runIt( void* arg)
{
#ifdef PTH_USE_TRACER
  InukTracer::openTrace();
#endif
  BasicThreadModel* s = (BasicThreadModel*)arg;
  void* retval = s->run();
#ifdef PTH_USE_TRACER 
  InukTracer::closeTrace();
#endif
  return retval;
};

inline int BasicThreadModel::startnothrow(Thread::Kind k) throw(PosixError)
{ return createnothrow(&_threadid, (void *(*)(void*))runIt, this , k) ; }

inline void BasicThreadModel::start(Thread::Kind k) throw(PosixError)
{ 
  //std::cout << "Start: @obj =" << (void*)this << std::endl;
  create(& _threadid, (void *(*)(void*))runIt , this, k ); 
}


inline CriticalGuard::CriticalGuard (Mutex &mtx) 
  : _mtx_(&mtx) 
{  _mtx_->lock();}
inline CriticalGuard::CriticalGuard (Mutex *mtx) 
  : _mtx_(mtx) 
{  _mtx_->lock();}
inline CriticalGuard::~CriticalGuard () 
{ _mtx_->unlock();}

template<class LOCKABLE>
inline TCriticalGuard<LOCKABLE>::TCriticalGuard (LOCKABLE &mtx) 
  : _mtx_(&mtx) 
{  _mtx_->lock();}

template<class LOCKABLE>
inline TCriticalGuard<LOCKABLE>::TCriticalGuard (LOCKABLE *mtx) 
  : _mtx_(mtx) 
{  _mtx_->lock();}

template<class LOCKABLE>
inline TCriticalGuard<LOCKABLE>::~TCriticalGuard () 
{ _mtx_->unlock();}


}// -namespace
#endif // _PTH_PTHREADSYNC_H_
