// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_pthreadsync_notrace.h,v $ 
//Copyright(c)'2001 by Apache Team // see the copyright file 
// $Id: pth_pthreadsync_notrace.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $ 
//==========================================================================

/*!
\file  
\brief  Posix thread for Inuktitut++ INRIA 
\author  J. Briat T. Gautier 
\date 2001

This file contains the wrappers to :
  - posix locks PTH::NTMutex
  - a lexical scoped lock mechanism CRITICAL()
  - thread specific data structure (Pth::NTKey)

This file contains wrappers of no_trace functions used by the tracer
mechanism of inuktitut. User level functions are in pth_pthreadsync.h
*/

#ifndef _PTH_PTHREADSYNC_NT_H
#define _PTH_PTHREADSYNC_NT_H

// the namespace identifier
#include <inuk/pth_ns.h>
#include <inuk/pth_dep.h>
#include <inuk/pth_macro.h>


#if defined(PTH_USE_PTHREAD_DARWIN) || defined(PTH_USE_PTHREAD_LINUX) || defined(PTH_USE_PTHREAD_IRIX) || defined(PTH_USE_PTHREAD_LINUX_IA64)
#include <pthread.h>
#endif 

namespace PTH { 

typedef pthread_mutex_t PMutex;
typedef pthread_cond_t  PCondition;

class Pth; // forward decl

//! @defgroup Synchro Synchronization facilities without trace
//! @brief All the classes to use for synchro.

// -------------------------------------------------------------------------
// class NTMutex
//! @brief Exclusive code execution.
//! @ingroup Synchro
// -------------------------------------------------------------------------

//!
//! This class implements interface for Posix mutex without trace or throwing
//! exception. Error code is returned to the caller.
//! @warning Be careful to unlock a mutex in case of exception throwing. 
//! @warning Use class PTH::CriticalGuard to do this safely.

class NTMutex {
public:
  //! Constructor. The mutex is NOT locked when returning.
  NTMutex()
#ifdef PTH_DEBUG
   : _isinit(false)  
#endif
  {}

  //! destructor
  ~NTMutex() {
#ifdef PTH_DEBUG
    _isinit =false;  
#endif
  } 
  
  //! Initialize a  (default case) mutex in unlock state.
  //! @param recursive if true the mutex is a recusrive one.
  //! @returns error code is !=0.
  int init()
  { return pthread_mutex_init(&_mutex,&default_attr); }
  
  //! Explicit destroy. Must be called before the destructor is called. 
  //! @returns error code is !=0.
  int destroy ()
  { return pthread_mutex_destroy(&_mutex); }

  //! Lock request. Wait until access is granted.
  //! @return 0 if OK, <>0 if the mutex is not initialized
  int lock()
  { return pthread_mutex_lock(&_mutex); }

  //! Unlock request.
  //! @return 0 if OK, <>0 if the mutex is not initialized or not owned by 
  //! the current thread.  
  int unlock()
  { return pthread_mutex_unlock(&_mutex); }

protected: 
  pthread_mutex_t _mutex;   //!< the posix lock
#ifdef PTH_DEBUG
  bool _isinit;
#endif
  static  pthread_mutexattr_t default_attr;
  static  pthread_mutexattr_t recursive_attr;

private :
  friend class Condition;
  friend class Pth;
};



// -------------------------------------------------------------------------
// class CriticalGuard
//! @brief Transparent mutual exclusion facility.
//! @ingroup Synchro
// -------------------------------------------------------------------------

//! This class implements a  transparent local scoping usage 
//! of mutex lock. An instance of this class is used to lock a block 
//! expression (i.e. between { and }).
//! The instance delaration must appear at the beginning of the block.
//! At the initialization of such an object, the associated mutex is locked 
//! and is unlocked at the destruction of the object (at the end of its 
//! scope as a C++ variable). Note that the unlock method is alway called 
//! even if an exception is thrown.
//! Find examples of use @ref pth_critical_ex "here".
//! @see CRITICAL
//! @see PTH::TCriticalGuard

class NTCriticalGuard {
private :
  PTH::NTMutex* _mtx_;
public : 
  NTCriticalGuard (NTMutex &mtx) : _mtx_(&mtx) {  _mtx_->lock();};
  NTCriticalGuard (NTMutex *mtx) : _mtx_(mtx) {  _mtx_->lock();};
  ~NTCriticalGuard () { _mtx_->unlock();};
};

//! @def  CRITICAL_NOTRACE()
//! @see PTH::NTCriticalGuard
#define CRITICAL_NOTRACE(MTX) PTH::NTCriticalGuard _CONCAT(_cg, __LINE__) (MTX)




// -------------------------------------------------------------------------
// class Key
//! Thread private attribute.
// -------------------------------------------------------------------------


//! This class provides an interface to posix pthread specific
//! attributes.
//! It only includes operators for :
//!  - getting specific data key create() with a value destructor function 
//!    called at the end of the thread
//!  - setting/getting specific data set() get()
template < class T> class NTKey {
protected :
  pthread_key_t key;
public :
  NTKey() throw()
  {}
  ~NTKey() throw()
  {}

  //! Allocate a key. 
  //! @throw PosixError if no more keys are available
  void create(void (*erase)(T*)=0) throw (PosixError)
  { int err; 
    ASSERT( !(err=pthread_key_create(&key,(void (*)(void*))erase)), 
      PosixError (SUFX("[NTKey::signal] POSIX error"),err));
  }

  //! Associate a thread specific value to a key. 
  //! @throw PosixError if the key is not correct.
  void set( T * v) throw (PosixError)
  { int err; 
    ASSERT( !(err=pthread_setspecific(key, (void *) v)), 
      PosixError (SUFX("[NTKey::signal] POSIX error"),err));
  }

  //! Get the thread specific value associated with a given key.
  //! @return the thread specific value of the key.
  T* get() throw ()
  { return (T*) pthread_getspecific(key); }
  

  //! @name Methods returning an error code instead of throwing exceptions
  //@{
  //! Allocate a key. 
  //! @return <>0 if no more keys are available.
  int createnothrow(void (*erase)(T*)=0) throw ()
  { return  pthread_key_create(&key,(void (*)(void*))erase ); }


  //! Associate a thread specific value to a key. 
  //! @return <>0 if the key is not correct.
  int setnothrow( T * v) throw ()
  { return pthread_setspecific(key, (void *) v); }
  //@}

};   




// -------------------------------------------------------------------------
// typedef NTThreadId
//! POSIX thread identifier
// -------------------------------------------------------------------------
typedef pthread_t NTThreadId;

} // namespace PTH

#ifdef PTH_USE_TRACER
#include <inuk/Pthread_stub_tracer.h>
#endif


#endif // _PTH_PTHREADSYNC_H_
