// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_staction.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// Description: stack object are used to setup correctly state of object
// in case of exception throwing. All constructions take two functions as
// input: begin is called at the construction phase of the object, while
// end is called at it destruction (may be due to exception).
// $Id: pth_staction.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================
#ifndef _PTH_STACTION_H
#define _PTH_STACTION_H

#include <inuk/pth_ns.h>
#include <inuk/pth_pthreadsync.h>

namespace PTH {

// -------------------------------------------------------------------------
/**StAction.
 StAction define stack object with two functions: the function begin and
the function end. The function begin is called at the construction
of the stack object and the function end at its destruction.
Such object can be use to provide an exception safe mechanism to 
ensure some execution arround a section of code.
From a certain point of view, this kind of object could be view as 
a basic to provide a Java like 'synchronized' method on object.
\verbatim


// -- Un-exception safe version -> f_at_the_end is never executed
// is an exception is thrown in do_main_compute.
void dummy1_toto() {
  // - do pre computation
  f_at_the_begining();

  // here do that can throw exception
  do_main_compute(); 

  // - do post computation
  f_at_the_end();
}

// -- Exception safe version -> f_at_the_end is always executed
// is an exception is thrown in do_main_compute, because the
// object stackobject is destroyed.
void dummy1_toto() {
  
  StAction stackobject( f_at_the_begining, f_at_the_end );

  // here do that can throw exception
  do_main_compute(); 
\endverbatim
}

*/
class StAction {
public:
  /** Constructor
   Construct a new stack object with an begin and end functions.
   The begin is called at the construction, the end function at
   the destruction of the stack object.
   Null parameter could be given, no action is done on it.
    @param begin the entry function
    @param end the exit function
  */
  StAction(void (*begin)(), void (*end)())
   : _end(end) {
    if (begin !=0) (*begin)();
  }
  /** Destructor
    Destroy the stack object, call the end function.  
   */
  ~StAction() { if (_end !=0) (*_end)(); }
  
private:
  void (*_end)();
};



// -------------------------------------------------------------------------
/** StObjAction
See StAction for a fine description of the purpose of stack objects.
A StObjAction stack object is provided in order to call methods on
an object both at the time of the construction and the destruction 
of the stack object, even if a exception is thrown
(StAction is just for function calls).
@see StAction
*/
template<class T>
class StObjAction {
public:
  /** Constructor
   Construct a new stack object with an begin and end methods.
   The begin method is called on the object ref at the 
   construction of the stack object, while the end method is called
   on the same object at the destruction of the stack object.
   Null parameter could be given, no action is done on it.
    @param ref the object on which method should be called.
    @param begin the entry method
    @param end the exit method
  */
  StObjAction( T& ref, void (T::*begin)(), void (T::*end)())
   : _refobj(ref), _end(end) {
    if (begin !=0) (_refobj.*begin)();
  }
  /** Destructor
    Destroy the stack object, call the end method on the object ref.  
   */
  ~StObjAction() { if (_end !=0) (_refobj.*_end)(); }
  
private:
  T& _refobj;
  void (T::*_end)();
};



// -------------------------------------------------------------------------
/**StPtrAction.
See StAction for a fine description of the purpose of stack objects.
A StPtrAction object is the same as an StObjAction except that 
it should be build using a pointer to an object in place of a reference.
@see StAction
@see StObjAction
*/
template<class T>
class StPtrAction {
public:
  /** Constructor
   Null parameter could be given, no action is done on it.
    @param pt the pointer object on which method should be called.
    @param begin the entry method
    @param end the exit method
  */
  StPtrAction( T* pt, void (T::*begin)(), void (T::*end)())
   : _ptobj(pt), _end(end) {
    if ((_ptobj !=0) && (begin !=0))  
      (_ptobj->*begin)();
  }
  /** Destructor
    Destroy the stack object, call the end method on the object ref.  
   */
  ~StPtrAction() { 
    if ((_ptobj !=0) && (_end !=0)) 
      (_ptobj->*_end)(); 
  }
  
private:
  T* _ptobj;
  void (T::*_end)();
};
// -------------------------------------------------------------------------
// Stack based mutex
/** StMutex.
This class implement a "stack mutex" used to 
define a critical section of code.
Such an object could be used to define a critical
section associated to the its C++ lexical scope
(i.e. between { and }).
At the initialization of such an object, the
associated mutex is locked and it is unlocked
at the destruction of the object (at the end 
of its scope as a C++ variable). 
Note that the unlock method is alway called 
even if an exception is thrown.
\verbatim

// - declaration but not initialization
static Mutex my_mutex;

void dummy1_toto() {
  // here do something

  { // I want this block to be executed in exclusive access
    StMutex critical (my_mutex);
    ....
    ....
  } // here the mutex is released

  // here do something
}
\endverbatim
*/
class StMutex : public StObjAction<Mutex> {
public:
  /// - Cstor: lock the allocated mutex
  StMutex(Mutex& mutex)
  : StObjAction<Mutex>(mutex, &Mutex::lock, &Mutex::unlock)
  {}
};


};

#endif // _PTH_STACTION_H_
