// ==========================================================================
// $Source
// Copyright(c)'2001 by APACHE team
// see the copyright file.
// Authors: T. Gautier
// [3/3/97]: based on givtimer.h from Givaro library.
//
//
//
// ==========================================================================
#ifndef _PTH_TIMER_H_
#define _PTH_TIMER_H_

#include <iostream>
#include <string>
#include <inuk/pth_dep.h>
#include <limits.h>
#ifdef INUKTITUT_USE_LINUX
//#include <asm/msr.h> // rdtsc
	#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
		 
#endif
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#ifdef INUKTITUT_USE_IRIX
#include <time.h> // clock
#endif


namespace PTH {

/* Forward declaration */
class Pth;

// -- 
// WallTimer class:
// --
class WallTimer {
public:
  /** */
  static const std::string& unit();

  typedef double type;

  /** */
  WallTimer( )
   : _delta(0) {}

  /** */
  void clear();

  /** */
  void start();

  /** */
  void stop();

  /** */
  double time() const;

  /** */
  static double gettime();
protected:
  type _delta;
};



// -- 
// CpuTimer class:
// --
class CpuTimer {
public:
  /** */
  static const std::string& unit();

  /** */
  typedef double type;

  /** */
  CpuTimer( )
   : _delta(0) {}

  /** */
  void clear();

  /** */
  void start();

  /** */
  void stop();

  /** */
  double time() const;

  /** */
  static double gettime();
protected:
  type _delta;
};



// -- 
// SysTimer class:
// --
class SysTimer {
public:
  /** */
  static const std::string& unit();

  /** */
  typedef double type;

  /** */
  SysTimer( )
   : _delta(0) {}

  /** */
  void clear();

  /** */
  void start();

  /** */
  void stop();

  /** */
  double time() const;

  /** */
  static double gettime();
protected:
  type _delta;
};


// -- 
// HighResolution timer class
// --
class HighResTimer {
public:
  /** */
  static const std::string& unit();

  /** */
  struct type { 
    type( int val =0 ) 
    { u.t =val; }
    type(unsigned long l) 
    { u.s.t_u =0; u.s.t_l =l; }
    type(unsigned long h, unsigned long l) 
    { u.s.t_u =h; u.s.t_l =l; }
    type(unsigned long long ll) 
    { u.t = ll; }
    operator double() const
    { return double(u.t); }
    operator long() const;
    operator unsigned long() const;
    operator unsigned long long() const;

    type& operator=(const type&);
    type& operator+=(const type& t);
    type& operator-=(const type& t);
    type operator-(const type& t) const;
    type operator+(const type& t) const;
    bool operator<(const type& t) const;
    bool operator>(const type& t) const;
    bool operator<(const unsigned int t) const;
    bool operator>(const unsigned int t) const;
    bool operator==(const type& t) const;
    bool operator==(const unsigned int t) const;

  public:
    union {
      struct {
        unsigned long t_l; 
        unsigned long t_u; 
      } s;
      unsigned long long t;
    } u;
  };

  /** */
  HighResTimer( )
   : _delta() {}

  /** */
  ~HighResTimer() {}

  /** */
  void clear();
  /** */
  void start();
  /** */
  void stop();

  /** */
  type tick() const
  { return _delta; }

  /** */
  double time() const
  { return double(_delta.u.t)/dtick_per_s; }

  /** */
  static type gettick();

  /** */
  static double gettime();

  /** number of tick per second */
  static double dtick_per_s;

protected:
  friend class type;
  friend class Pth;
  static void calibrate();
  static unsigned long long latency;
#ifdef INUKTITUT_USE_IRIX
  static struct timespec tspec;
#endif
  type  _delta;
};

// -- 
// LogicalTimer class:
// --
class LogicalTimer {
public:
  /** */
  static const std::string& unit();

  /** */
  typedef unsigned short type;

  /** */
  LogicalTimer( )
   : _delta(0) {}

  /** */
  void clear();

  /** */
  void start();

  /** */
  void stop();

  /** */
  type tick() const;

  /** */
  double time() const;

protected:
  type _delta;
};


// -- 
// Timer class: contains three objects of the previous type
// --
/** Timer.
Class that defines stop/start timers to get 
timing informations (CPU time, System time, Wall time or wall clock).
The value that stores a timer is defined after a call to stop method
and represents the delay (interval of time) between the calls to start
and stop.
@author T. Gautier
@version $Id: pth_timer.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
*/
class Timer : protected HighResTimer, 
              protected WallTimer,
              protected CpuTimer,
              protected SysTimer,
              protected LogicalTimer {
public :

  /** Type for timer
   */
  struct type {
    type& operator-=( const type& a );
    type& operator+=( const type& a );
    HighResTimer::type t_hr;
    WallTimer::type    t_r;
    CpuTimer::type     t_u;
    SysTimer::type     t_s;
    LogicalTimer::type t_l;
  };

  /** Clear timers. Reset internal value of the object.
  */ 
  void clear(); 

  /** Start timers. 
  */
  void start();

  /** Stop timers.
     The different times spent between the start and the stop method calls
     is given by methods usertime, systime, realtime.
  */
  void stop();

  /** Return the total amount of second spent in user mode
    (CPU time).
  */
  double usertime() const;

  /** Return the total amount of second spent in system mode
    (System time)
  */
  double systime () const;

  /** Return the total amount of second spent by
    a wall clock (or real time). High Resolution timer
  */
  double hrrealtime () const;

  /** Return the total amount of second spent by
    a wall clock (or real time).
  */
  double realtime () const;

  /** Return the number of clock tick.
  */
  double tickcount () const;

  /** Return the number of start/stop.
  */
  double callcount () const;

  /** 
     Return a seed value to initialize random generator.
  */
  static long seed();

  /**
    Output to a stream. 
    operator<< is also defined.
  */
  std::ostream& print( std::ostream& ) const;

private:
};



// -- 
// Timer class: contains three objects of the previous type
// -- 
template<class T1,class T2>
class PairTimer : public T1, public T2 {
public :
  /** */
  static const std::string& unit();

  /** Type for timer
   */
  struct type {
    type& operator-=( const type& a )
    {
      t_1 -= a.t_1;
      t_2 -= a.t_2;
      return *this;
    }
    type& operator+=( const type& a )
    {
      t_1 += a.t_1;
      t_2 += a.t_2;
      return *this;
    }
    typename T1::type t_1;
    typename T2::type t_2;
  };

  /** */
  void clear()
  { T1::clear(); T2::clear(); }

  /** */
  void start()
  { T1::start(); T2::start(); }

  /** */
  void stop()
  { T1::stop(); T2::stop(); }

  /** */
  double time1() const 
  { return T1::time(); }

  /** */
  double time2() const 
  { return T2::time(); }

  /** */
  std::ostream& print( std::ostream& ) const;

private:
};


/*
 * inline definitions
 */
inline HighResTimer::type& HighResTimer::type::operator+=(const HighResTimer::type& t) 
{ u.t += t.u.t; return *this;  }

inline HighResTimer::type& HighResTimer::type::operator-=(const HighResTimer::type& t) 
{ u.t -= t.u.t; return *this; }

inline HighResTimer::type::operator long() const
{ return u.t; }

inline HighResTimer::type::operator unsigned long() const
{ return u.t; }

inline HighResTimer::type::operator unsigned long long() const
{ return u.t; }

inline HighResTimer::type HighResTimer::type::operator-(const HighResTimer::type& t) const
{ return u.t - t.u.t; }

inline HighResTimer::type HighResTimer::type::operator+(const HighResTimer::type& t) const
{ return u.t + t.u.t; }

inline bool HighResTimer::type::operator<(const HighResTimer::type& t) const
{ return u.t < t.u.t; }

inline bool HighResTimer::type::operator>(const HighResTimer::type& t) const
{ return u.t > t.u.t; }

inline bool HighResTimer::type::operator<(const unsigned int t) const
{ return u.t < t; }

inline bool HighResTimer::type::operator>(const unsigned int t) const
{ return u.t > t; }

inline bool HighResTimer::type::operator==(const HighResTimer::type& t) const
{ return u.t == t.u.t; }

inline bool HighResTimer::type::operator==(const unsigned int t) const
{ return u.t == t; }

inline HighResTimer::type& HighResTimer::type::operator=(const HighResTimer::type& v)
{ u.t = v.u.t; return *this; }

inline HighResTimer::type HighResTimer::gettick()
{
#if defined(INUKTITUT_USE_DARWIN) || defined(INUKTITUT_USE_LINUX)
#ifdef __ppc__
  register unsigned long t_u;
  register unsigned long t_l;
  asm volatile ("mftbu %0" : "=r" (t_u) );
  asm volatile ("mftb %0" : "=r" (t_l) );
  return type(t_u,t_l);
#elif defined(__x86__)
  register unsigned long t_u;
  register unsigned long t_l;
  rdtsc( t_l, t_u );
  return type(t_u,t_l);
#endif
#elif defined(INUKTITUT_USE_IRIX)
  struct timespec tp;
  clock_gettime( CLOCK_SGI_CYCLE, &tp );
  return type((unsigned long long) (double(tp.tv_sec)  + double(tp.tv_nsec)/1e9) );
#endif
}

inline void HighResTimer::clear()
{ _delta.u.t =0; }

inline void HighResTimer::start()
{ _delta = HighResTimer::gettick(); }

inline void HighResTimer::stop()
{ _delta.u.t = HighResTimer::gettick().u.t - _delta.u.t - HighResTimer::latency; }

inline double HighResTimer::gettime()
{ return double(HighResTimer::gettick().u.t)/dtick_per_s; }

inline void WallTimer::clear()
{ _delta =0; }

inline double WallTimer::gettime()
{ 
   struct timeval tmp2 ;
   gettimeofday (&tmp2, 0) ;

   // real time 
   return (double) tmp2.tv_sec + ((double) tmp2.tv_usec) * 1e-6;
}

inline void WallTimer::start()
{ 
  _delta = WallTimer::gettime();
}

inline void WallTimer::stop()
{ 
  _delta = WallTimer::gettime() - _delta;
}

inline double WallTimer::time() const
{ 
  return _delta;
}


inline void CpuTimer::clear()
{ _delta =0; }

inline double CpuTimer::gettime()
{
   struct rusage tmp;
   getrusage (RUSAGE_SELF, &tmp) ;
   // user time
   return double(tmp.ru_utime.tv_sec) + double(tmp.ru_utime.tv_usec) * 1e-6;
}

inline void CpuTimer::start()
{
  _delta = CpuTimer::gettime();
}

inline void CpuTimer::stop()
{
  _delta = CpuTimer::gettime() - _delta;
}

inline double CpuTimer::time() const
{
  return _delta;
}


inline void SysTimer::clear()
{ _delta =0; }

inline double SysTimer::gettime()
{
   struct rusage tmp;
   getrusage (RUSAGE_SELF, &tmp) ;
   // user time
   return double(tmp.ru_stime.tv_sec) + double(tmp.ru_stime.tv_usec) * 1e-6;
}

inline void SysTimer::start()
{
  _delta = SysTimer::gettime();
}

inline void SysTimer::stop()
{
  _delta = SysTimer::gettime() - _delta;
}

inline double SysTimer::time() const
{
  return _delta;
}

 
inline void LogicalTimer::clear()
{ _delta =0; }

inline void LogicalTimer::start()
{
  _delta =0;
}

inline void LogicalTimer::stop()
{
  _delta =1;
}

inline LogicalTimer::type LogicalTimer::tick() const
{
  return _delta;
}

inline double LogicalTimer::time() const
{
  return double(_delta);
}

inline std::ostream& operator<<( std::ostream& o, const Timer& T)
{ return T.print(o);}

inline Timer::type& Timer::type::operator-=( const Timer::type& a )
{
  t_hr -= a.t_hr;
  t_r -= a.t_r;
  t_u -= a.t_u;
  t_s -= a.t_s;
  t_l -= a.t_l;
  return *this;
}

inline Timer::type& Timer::type::operator+=( const Timer::type& a )
{
  t_hr += a.t_hr;
  t_r += a.t_r;
  t_u += a.t_u;
  t_s += a.t_s;
  t_l += a.t_l;
  return *this;
}


inline void Timer::clear()
{ 
  HighResTimer::clear();
  WallTimer::clear();
  CpuTimer::clear();
  SysTimer::clear();
  LogicalTimer::clear();
}

inline void Timer::start()
{
  HighResTimer::start();
  WallTimer::start();
  CpuTimer::start();
  SysTimer::start();
  LogicalTimer::start();
}

inline void Timer::stop()
{
  HighResTimer::stop();
  WallTimer::stop();
  CpuTimer::stop();
  SysTimer::stop();
  LogicalTimer::stop();
}

inline double Timer::usertime() const 
{ return CpuTimer::time(); }

inline double Timer::systime () const 
{ return SysTimer::time(); }

inline double Timer::hrrealtime () const 
{ return HighResTimer::time(); }

inline double Timer::realtime () const 
{ return WallTimer::time(); }

inline double Timer::tickcount () const 
{ return HighResTimer::tick(); }

inline double Timer::callcount () const 
{ return LogicalTimer::time(); }


template<class T1, class T2>
inline std::ostream& operator<<( std::ostream& o, const PairTimer<T1,T2>& T )
{ return T.print(o);}

template<class T1, class T2>
const std::string& PairTimer<T1,T2>::unit()
{
  static std::string u = std::string("[") + T1::unit() + "," + T2::unit() + std::string("]");
  return u;
}

template<class T1, class T2>
std::ostream& PairTimer<T1,T2>::print( std::ostream& o ) const
{
  o << '[' << time1() << ',' << time2() << ']';
  return o;
}

} // - namespace


#endif 
